from __future__ import annotations

import copy
import warnings
from builtins import isinstance
from typing import (
    TYPE_CHECKING,
    Any,
    Iterable,
    Optional,
    Sequence,
    Type,
    TypeVar,
    Union,
    cast,
)

import numpy as np
import pandas.api.extensions

from skfda._utils._utils import _to_array_maybe_ragged

from ..._utils import _check_array_key, _int_to_real, constants
from .. import grid
from .._functional_data import FData
from .._typing import (
    ArrayLike,
    DomainRange,
    GridPointsLike,
    LabelTupleLike,
    NDArrayInt,
)
from ..extrapolation import ExtrapolationLike
from . import Basis

if TYPE_CHECKING:
    from .. import FDataGrid

T = TypeVar('T', bound='FDataBasis')


class FDataBasis(FData):  # noqa: WPS214
    r"""Basis representation of functional data.

    Class representation for functional data in the form of a set of basis
    functions multplied by a set of coefficients.

    .. math::
        f(x) = \sum_{k=1}{K}c_k\phi_k

    Where n is the number of basis functions, :math:`c = (c_1, c_2, ...,
    c_K)` the vector of coefficients and  :math:`\phi = (\phi_1, \phi_2,
    ..., \phi_K)` the basis function system.

    Attributes:
        basis: Basis function system.
        coefficients: List or matrix of coefficients. Has to
            have the same length or number of columns as the number of basis
            function in the basis. If a matrix, each row contains the
            coefficients that multiplied by the basis functions produce each
            functional datum.
        domain_range: 2 dimension matrix where each row
            contains the bounds of the interval in which the functional data
            is considered to exist for each one of the axies.
        dataset_name: name of the dataset.
        argument_names: tuple containing the names of the different
            arguments.
        coordinate_names: tuple containing the names of the different
            coordinate functions.
        extrapolation: defines the default type of
            extrapolation. By default None, which does not apply any type of
            extrapolation. See `Extrapolation` for detailled information of the
            types of extrapolation.

    Examples:
        >>> from skfda.representation.basis import FDataBasis, Monomial
        >>>
        >>> basis = Monomial(n_basis=4)
        >>> coefficients = [1, 1, 3, .5]
        >>> FDataBasis(basis, coefficients)
        FDataBasis(
            basis=Monomial(domain_range=((0, 1),), n_basis=4),
            coefficients=[[ 1.   1.   3.   0.5]],
            ...)

    """

    def __init__(
        self,
        basis: Basis,
        coefficients: ArrayLike,
        *,
        dataset_label: Optional[str] = None,
        dataset_name: Optional[str] = None,
        axes_labels: Optional[LabelTupleLike] = None,
        argument_names: Optional[LabelTupleLike] = None,
        coordinate_names: Optional[LabelTupleLike] = None,
        sample_names: Optional[LabelTupleLike] = None,
        extrapolation: Optional[ExtrapolationLike] = None,
    ) -> None:
        """Construct a FDataBasis object."""
        coefficients = _int_to_real(np.atleast_2d(coefficients))
        if coefficients.shape[1] != basis.n_basis:
            raise ValueError(
                "The length or number of columns of coefficients "
                "has to be the same equal to the number of "
                "elements of the basis.",
            )
        self.basis = basis
        self.coefficients = coefficients

        super().__init__(
            extrapolation=extrapolation,
            dataset_label=dataset_label,
            dataset_name=dataset_name,
            axes_labels=axes_labels,
            argument_names=argument_names,
            coordinate_names=coordinate_names,
            sample_names=sample_names,
        )

    @classmethod
    def from_data(
        cls,
        data_matrix: np.ndarray,
        *,
        basis: Basis,
        grid_points: Optional[GridPointsLike] = None,
        sample_points: Optional[GridPointsLike] = None,
        method: str = 'cholesky',
    ) -> FDataBasis:
        r"""Transform raw data to a smooth functional form.

        Takes functional data in a discrete form and makes an approximates it
        to the closest function that can be generated by the basis. This
        function does not attempt to smooth the original data. If smoothing
        is desired, it is better to use :class:`BasisSmoother`.

        The fit is made so as to reduce the sum of squared errors
        [RS05-5-2-5]_:

        .. math::

            SSE(c) = (y - \Phi c)' (y - \Phi c)

        where :math:`y` is the vector or matrix of observations, :math:`\Phi`
        the matrix whose columns are the basis functions evaluated at the
        sampling points and :math:`c` the coefficient vector or matrix to be
        estimated.

        By deriving the first formula we obtain the closed formed of the
        estimated coefficients matrix:

        .. math::

            \hat{c} = \left( \Phi' \Phi \right)^{-1} \Phi' y

        The solution of this matrix equation is done using the cholesky
        method for the resolution of a LS problem. If this method throughs a
        rounding error warning you may want to use the QR factorisation that
        is more numerically stable despite being more expensive to compute.
        [RS05-5-2-7]_

        Args:
            data_matrix: List or matrix containing the
                observations. If a matrix each row represents a single
                functional datum and the columns the different observations.
            grid_points: Values of the domain where the previous
                data were taken.
            basis: Basis used.
            method: Algorithm used for calculating the coefficients using
                the least squares method. The values admitted are 'cholesky'
                and 'qr' for Cholesky and QR factorisation methods
                respectively.
            sample_points: Old name for `grid_points`. New code should
                use `grid_points` instead.

                .. deprecated:: 0.5

        Returns:
            FDataBasis: Represention of the data in a functional form as
                product of coefficients by basis functions.

        Examples:
            >>> import numpy as np
            >>> t = np.linspace(0, 1, 5)
            >>> x = np.sin(2 * np.pi * t) + np.cos(2 * np.pi * t) + 2
            >>> x
            array([ 3.,  3.,  1.,  1.,  3.])

            >>> from skfda.representation.basis import FDataBasis, Fourier
            >>> basis = Fourier((0, 1), n_basis=3)
            >>> fd = FDataBasis.from_data(x, grid_points=t, basis=basis)
            >>> fd.coefficients.round(2)
            array([[ 2.  , 0.71, 0.71]])

        References:
            .. [RS05-5-2-5] Ramsay, J., Silverman, B. W. (2005). How spline
                smooths are computed. In *Functional Data Analysis*
                (pp. 86-87). Springer.

            .. [RS05-5-2-7] Ramsay, J., Silverman, B. W. (2005). HSpline
                smoothing as an augmented least squares problem. In *Functional
                Data Analysis* (pp. 86-87). Springer.

        """
        if sample_points is not None:
            warnings.warn(
                "Parameter sample_points is deprecated. Use the "
                "parameter grid_points instead.",
                DeprecationWarning,
            )
            grid_points = sample_points

        # n is the samples
        # m is the observations
        # k is the number of elements of the basis

        # Each sample in a column (m x n)
        data_matrix = np.atleast_2d(data_matrix)

        fd = grid.FDataGrid(data_matrix=data_matrix, grid_points=grid_points)

        return fd.to_basis(basis=basis, method=method)

    @property
    def n_samples(self) -> int:
        return len(self.coefficients)

    @property
    def dim_domain(self) -> int:
        return self.basis.dim_domain

    @property
    def dim_codomain(self) -> int:
        return self.basis.dim_codomain

    @property
    def coordinates(self: T) -> _CoordinateIterator[T]:
        r"""Return a component of the FDataBasis.

        If the functional object contains samples
        :math:`f: \mathbb{R}^n \rightarrow \mathbb{R}^d`, this object allows
        a component of the vector :math:`f = (f_1, ..., f_d)`.

        """
        return _CoordinateIterator(self)

    @property
    def n_basis(self) -> int:
        """Return number of basis."""
        return self.basis.n_basis

    @property
    def domain_range(self) -> DomainRange:

        return self.basis.domain_range

    def _evaluate(
        self,
        eval_points: Union[ArrayLike, Iterable[ArrayLike]],
        *,
        aligned: bool = True,
    ) -> np.ndarray:

        if aligned:

            eval_points = np.asarray(eval_points)

            # Each row contains the values of one element of the basis
            basis_values = self.basis.evaluate(eval_points)

            res = np.tensordot(self.coefficients, basis_values, axes=(1, 0))

            return res.reshape(
                (self.n_samples, len(eval_points), self.dim_codomain),
            )

        eval_points = cast(Iterable[ArrayLike], eval_points)

        res_list = [
            np.sum((c * self.basis.evaluate(np.asarray(p)).T).T, axis=0)
            for c, p in zip(self.coefficients, eval_points)
        ]

        return _to_array_maybe_ragged(res_list)

    def shift(
        self,
        shifts: Union[ArrayLike, float],
        *,
        restrict_domain: bool = False,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid_points: Optional[GridPointsLike] = None,
    ) -> FDataGrid:
        r"""
        Perform a shift of the curves.

        The i-th shifted function :math:`y_i` has the form

        .. math::
            y_i(t) = x_i(t + \delta_i)

        where :math:`x_i` is the i-th original function and :math:`delta_i` is
        the shift performed for that function, that must be a vector in the
        domain space.

        Note that a positive shift moves the graph of the function in the
        negative direction and vice versa.

        Args:
            shifts: List with the shifts
                corresponding for each sample or numeric with the shift to
                apply to all samples.
            restrict_domain: If True restricts the domain to avoid the
                evaluation of points outside the domain using extrapolation.
                Defaults uses extrapolation.
            extrapolation: Controls the
                extrapolation mode for elements outside the domain range.
                By default uses the method defined in fd. See extrapolation to
                more information.
            grid_points: Grid of points where
                the functions are evaluated to obtain the discrete
                representation of the object to operate. If ``None`` the
                current grid_points are used to unificate the domain of the
                shifted data.

        Returns:
            Shifted functions.

        """
        grid_points = (
            self._default_grid_points() if grid_points is None
            else grid_points
        )

        return super().shift(
            shifts=shifts,
            restrict_domain=restrict_domain,
            extrapolation=extrapolation,
            grid_points=grid_points,
        )

    def derivative(self: T, *, order: int = 1) -> T:  # noqa: D102

        if order < 0:
            raise ValueError("order only takes non-negative integer values.")

        if order == 0:
            return self.copy()

        basis, coefficients = self.basis.derivative_basis_and_coefs(
            self.coefficients,
            order,
        )

        return self.copy(basis=basis, coefficients=coefficients)

    def sum(  # noqa: WPS125
        self: T,
        *,
        axis: Optional[int] = None,
        out: None = None,
        keepdims: bool = False,
        skipna: bool = False,
        min_count: int = 0,
    ) -> T:
        """Compute the sum of all the samples in a FDataBasis object.

        Args:
            axis: Used for compatibility with numpy. Must be None or 0.
            out: Used for compatibility with numpy. Must be None.
            keepdims: Used for compatibility with numpy. Must be False.
            skipna: Wether the NaNs are ignored or not.
            min_count: Number of valid (non NaN) data to have in order
                for the a variable to not be NaN when `skipna` is
                `True`.

        Returns:
            A FDataBais object with just one sample
            representing the sum of all the samples in the original
            FDataBasis object.

        Examples:
            >>> from skfda.representation.basis import FDataBasis, Monomial
            >>> basis = Monomial(n_basis=4)
            >>> coefficients = [[0.5, 1, 2, .5], [1.5, 1, 4, .5]]
            >>> FDataBasis(basis, coefficients).sum()
            FDataBasis(
                basis=Monomial(domain_range=((0, 1),), n_basis=4),
                coefficients=[[ 2.  2.  6.  1.]],
                ...)

        """
        super().sum(axis=axis, out=out, keepdims=keepdims, skipna=skipna)

        coefs = (
            np.nansum(self.coefficients, axis=0) if skipna
            else np.sum(self.coefficients, axis=0)
        )

        if min_count > 0:
            valid = ~np.isnan(self.coefficients)
            n_valid = np.sum(valid, axis=0)
            coefs[n_valid < min_count] = np.NaN

        return self.copy(
            coefficients=coefs,
            sample_names=(None,),
        )

    def gmean(self: T, eval_points: Optional[np.ndarray] = None) -> T:
        """Compute the geometric mean of the functional data object.

        A numerical approach its used. The object its transformed into its
        discrete representation and then the geometric mean is computed and
        then the object is taken back to the basis representation.

        Args:
            eval_points: Set of points where the
                functions are evaluated to obtain the discrete
                representation of the object. If none are passed it calls
                numpy.linspace with bounds equal to the ones defined in
                self.domain_range and the number of points the maximum
                between 501 and 10 times the number of basis.

        Returns:
            Geometric mean of the original object.

        """
        return self.to_grid(eval_points).gmean().to_basis(self.basis)

    def var(self: T, eval_points: Optional[np.ndarray] = None) -> T:
        """Compute the variance of the functional data object.

        A numerical approach its used. The object its transformed into its
        discrete representation and then the variance is computed and
        then the object is taken back to the basis representation.

        Args:
            eval_points: Set of points where the
                functions are evaluated to obtain the discrete
                representation of the object. If none are passed it calls
                numpy.linspace with bounds equal to the ones defined in
                self.domain_range and the number of points the maximum
                between 501 and 10 times the number of basis.

        Returns:
            Variance of the original object.

        """
        return self.to_grid(eval_points).var().to_basis(self.basis)

    def cov(self, eval_points: Optional[np.ndarray] = None) -> FData:
        """Compute the covariance of the functional data object.

        A numerical approach its used. The object its transformed into its
        discrete representation and then the covariance matrix is computed.

        Args:
            eval_points: Set of points where the
                functions are evaluated to obtain the discrete
                representation of the object. If none are passed it calls
                numpy.linspace with bounds equal to the ones defined in
                self.domain_range and the number of points the maximum
                between 501 and 10 times the number of basis.

        Returns:
            Matrix of covariances.

        """
        return self.to_grid(eval_points).cov()

    def to_grid(
        self,
        grid_points: Optional[GridPointsLike] = None,
        *,
        sample_points: Optional[GridPointsLike] = None,
    ) -> FDataGrid:
        """Return the discrete representation of the object.

        Args:
            grid_points (array_like, optional): Points per axis where the
                functions are evaluated. If none are passed it calls
                numpy.linspace with bounds equal to the ones defined in
                self.domain_range and the number of points the maximum
                between 501 and 10 times the number of basis.
            sample_points: Old name for `grid_points`. New code should
                use `grid_points` instead.

                .. deprecated:: 0.5

        Returns:
              FDataGrid: Discrete representation of the functional data
              object.

        Examples:
            >>> from skfda.representation.basis import FDataBasis, Monomial
            >>> fd = FDataBasis(coefficients=[[1, 1, 1], [1, 0, 1]],
            ...                 basis=Monomial(domain_range=(0,5), n_basis=3))
            >>> fd.to_grid([0, 1, 2])
            FDataGrid(
                array([[[ 1.],
                        [ 3.],
                        [ 7.]],
                       [[ 1.],
                        [ 2.],
                        [ 5.]]]),
                grid_points=(array([ 0., 1., 2.]),),
                domain_range=((0, 5),),
                ...)

        """
        if sample_points is not None:
            warnings.warn(
                "Parameter sample_points is deprecated. Use the "
                "parameter grid_points instead.",
                DeprecationWarning,
            )
            grid_points = sample_points

        if grid_points is None:
            grid_points = self._default_grid_points()

        return grid.FDataGrid(
            self.evaluate(grid_points, grid=True),
            grid_points=grid_points,
            domain_range=self.domain_range,
        )

    def to_basis(
        self,
        basis: Optional[Basis] = None,
        eval_points: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> FDataBasis:
        """
        Return the basis representation of the object.

        Args:
            basis: Basis object in which the functional data are
                going to be represented.
            eval_points: Evaluation points used to discretize the function
                if the basis is going to be changed.
            kwargs: Keyword arguments to be passed to
                FDataBasis.from_data().

        Returns:
            Basis representation of the funtional data object.

        """
        if basis is None or basis == self.basis:
            return self.copy()

        return self.to_grid(grid_points=eval_points).to_basis(basis, **kwargs)

    def copy(
        self: T,
        *,
        deep: bool = False,  # For Pandas compatibility
        basis: Optional[Basis] = None,
        coefficients: Optional[np.ndarray] = None,
        dataset_name: Optional[str] = None,
        argument_names: Optional[LabelTupleLike] = None,
        coordinate_names: Optional[LabelTupleLike] = None,
        sample_names: Optional[LabelTupleLike] = None,
        extrapolation: Optional[ExtrapolationLike] = None,
    ) -> T:
        """Copy the FDataBasis."""
        if basis is None:
            basis = copy.deepcopy(self.basis)

        if coefficients is None:
            coefficients = self.coefficients

        if dataset_name is None:
            dataset_name = self.dataset_name

        if argument_names is None:
            # Tuple, immutable
            argument_names = self.argument_names

        if coordinate_names is None:
            # Tuple, immutable
            coordinate_names = self.coordinate_names

        if sample_names is None:
            # Tuple, immutable
            sample_names = self.sample_names

        if extrapolation is None:
            extrapolation = self.extrapolation

        return FDataBasis(
            basis,
            coefficients,
            dataset_name=dataset_name,
            argument_names=argument_names,
            coordinate_names=coordinate_names,
            sample_names=sample_names,
            extrapolation=extrapolation,
        )

    def _default_grid_points(self) -> GridPointsLike:
        npoints = constants.N_POINTS_FINE_MESH
        return [
            np.linspace(*r, npoints)
            for r in self.domain_range
        ]

    def _to_R(self) -> str:  # noqa: N802
        """Return the code to build the object on fda package on R."""
        return (
            f"fd("  # noqa: WPS437
            f"coef = {self._array_to_R(self.coefficients, transpose=True)},"
            f" basisobj = {self.basis._to_R()})"
        )

    def _array_to_R(  # noqa: N802
        self,
        coefficients: np.ndarray,
        transpose: bool = False,
    ) -> str:
        if len(coefficients.shape) == 1:
            coefficients = coefficients.reshape((1, coefficients.shape[0]))

        if transpose is True:
            coefficients = np.transpose(coefficients)

        (rows, cols) = coefficients.shape
        retstring = "matrix(c("
        retstring += "".join(
            f"{coefficients[i, j]}, "
            for j in range(cols)
            for i in range(rows)
        )

        return (
            retstring[:len(retstring) - 2]
            + f"), nrow = {rows}, ncol = {cols})"
        )

    def __repr__(self) -> str:

        return (
            f"{self.__class__.__name__}("  # noqa: WPS221
            f"\nbasis={self.basis},"
            f"\ncoefficients={self.coefficients},"
            f"\ndataset_name={self.dataset_name},"
            f"\nargument_names={repr(self.argument_names)},"
            f"\ncoordinate_names={repr(self.coordinate_names)},"
            f"\nextrapolation={self.extrapolation})"
        ).replace('\n', '\n    ')

    def __str__(self) -> str:

        return (
            f"{self.__class__.__name__}("
            f"\n_basis={self.basis},"
            f"\ncoefficients={self.coefficients})"
        ).replace('\n', '\n    ')

    def equals(self, other: object) -> bool:
        """Equality of FDataBasis."""
        # TODO check all other params

        if not super().equals(other):
            return False

        other = cast(grid.FDataGrid, other)

        return (
            self.basis == other.basis
            and np.array_equal(self.coefficients, other.coefficients)
        )

    def __eq__(self, other: object) -> np.ndarray:  # type: ignore[override]
        """Elementwise equality of FDataBasis."""
        if not isinstance(other, type(self)) or self.dtype != other.dtype:
            if other is pandas.NA:
                return self.isna()
            if pandas.api.types.is_list_like(other) and not isinstance(
                other, (pandas.Series, pandas.Index, pandas.DataFrame),
            ):
                return np.concatenate([x == y for x, y in zip(self, other)])

            return NotImplemented

        if len(self) != len(other) and len(self) != 1 and len(other) != 1:
            raise ValueError(
                f"Different lengths: "
                f"len(self)={len(self)} and "
                f"len(other)={len(other)}",
            )

        return np.all(self.coefficients == other.coefficients, axis=1)

    def concatenate(
        self: T,
        *others: T,
        as_coordinates: bool = False,
    ) -> T:
        """
        Join samples from a similar FDataBasis object.

        Joins samples from another FDataBasis object if they have the same
        basis.

        Args:
            others: Objects to be concatenated.
            as_coordinates:  If False concatenates as
                new samples, else, concatenates the other functions as
                new components of the image. Defaults to False.

        Returns:
            :class:`FDataBasis`: FDataBasis object with the samples from the
            original objects.

        Todo:
            By the moment, only unidimensional objects are supported in basis
            representation.

        """
        # TODO: Change to support multivariate functions
        #  in basis representation
        if as_coordinates:
            return NotImplemented

        for other in others:
            if other.basis != self.basis:
                raise ValueError("The objects should have the same basis.")

        data = [self.coefficients] + [other.coefficients for other in others]

        sample_names = [fd.sample_names for fd in (self, *others)]

        return self.copy(
            coefficients=np.concatenate(data, axis=0),
            sample_names=sum(sample_names, ()),
        )

    def compose(
        self,
        fd: FData,
        *,
        eval_points: Optional[np.ndarray] = None,
        **kwargs: Any,
    ) -> FData:
        """
        Composition of functions.

        Performs the composition of functions. The basis is discretized to
        compute the composition.

        Args:
            fd (:class:`FData`): FData object to make the composition. Should
                have the same number of samples and image dimension equal to 1.
            eval_points (array_like): Points to perform the evaluation.
             kwargs: Named arguments to be passed to :func:`from_data`.

        Returns:
            Function resulting from the composition.

        """
        fd_grid = self.to_grid().compose(fd, eval_points=eval_points)

        if fd.dim_domain == 1:
            basis = self.basis.rescale(fd.domain_range[0])
            composition = fd_grid.to_basis(basis, **kwargs)
        else:
            #  Cant be convertered to basis due to the dimensions
            composition = fd_grid

        return composition

    def __getitem__(self: T, key: Union[int, slice, np.ndarray]) -> T:
        """Return self[key]."""
        key = _check_array_key(self.coefficients, key)

        return self.copy(
            coefficients=self.coefficients[key],
            sample_names=np.array(self.sample_names)[key],
        )

    def __add__(self: T, other: Union[T, np.ndarray, float]) -> T:
        """Addition for FDataBasis object."""
        if isinstance(other, FDataBasis) and self.basis == other.basis:

            return self._copy_op(
                other,
                basis=self.basis,
                coefficients=self.coefficients + other.coefficients,
            )

        return NotImplemented

    def __radd__(self: T, other: Union[T, np.ndarray, float]) -> T:
        """Addition for FDataBasis object."""
        if isinstance(other, FDataBasis) and self.basis == other.basis:

            return self._copy_op(
                other,
                basis=self.basis,
                coefficients=self.coefficients + other.coefficients,
            )

        return NotImplemented

    def __sub__(self: T, other: Union[T, np.ndarray, float]) -> T:
        """Subtraction for FDataBasis object."""
        if isinstance(other, FDataBasis) and self.basis == other.basis:

            return self._copy_op(
                other,
                basis=self.basis,
                coefficients=self.coefficients - other.coefficients,
            )

        return NotImplemented

    def __rsub__(self: T, other: Union[T, np.ndarray, float]) -> T:
        """Right subtraction for FDataBasis object."""
        if isinstance(other, FDataBasis) and self.basis == other.basis:

            return self._copy_op(
                other,
                basis=self.basis,
                coefficients=other.coefficients - self.coefficients,
            )

        return NotImplemented

    def _mul_scalar(self: T, other: Union[np.ndarray, float]) -> T:
        """Multiplication by scalar."""
        try:
            vector = np.atleast_1d(other)
        except Exception:
            return NotImplemented

        if vector.ndim > 1:
            return NotImplemented

        vector = vector[:, np.newaxis]

        return self._copy_op(
            other,
            basis=self.basis,
            coefficients=self.coefficients * vector,
        )

    def __mul__(self: T, other: Union[np.ndarray, float]) -> T:
        """Multiplication for FDataBasis object."""
        return self._mul_scalar(other)

    def __rmul__(self: T, other: Union[np.ndarray, float]) -> T:
        """Multiplication for FDataBasis object."""
        return self._mul_scalar(other)

    def __truediv__(self: T, other: Union[np.ndarray, float]) -> T:
        """Division for FDataBasis object."""
        try:
            other = 1 / np.asarray(other)
        except Exception:
            return NotImplemented

        return self._mul_scalar(other)

    def __rtruediv__(self: T, other: Union[np.ndarray, float]) -> T:
        """Right division for FDataBasis object."""
        return NotImplemented

    #####################################################################
    # Pandas ExtensionArray methods
    #####################################################################
    def _take_allow_fill(
        self: T,
        indices: NDArrayInt,
        fill_value: T,
    ) -> T:
        result = self.copy()
        result.coefficients = np.full(
            (len(indices),) + self.coefficients.shape[1:],
            np.nan,
        )

        positive_mask = indices >= 0
        result.coefficients[positive_mask] = self.coefficients[
            indices[positive_mask]
        ]

        if fill_value is not self.dtype.na_value:
            result.coefficients[~positive_mask] = fill_value.coefficients[0]

        return result

    @property
    def dtype(self) -> FDataBasisDType:
        """The dtype for this extension array, FDataGridDType"""
        return FDataBasisDType(basis=self.basis)

    @property
    def nbytes(self) -> int:
        """
        The number of bytes needed to store this object in memory.
        """
        return self.coefficients.nbytes

    def isna(self) -> np.ndarray:
        """
        Return a 1-D array indicating if each value is missing.

        Returns:
            na_values (np.ndarray): Positions of NA.
        """
        return np.all(np.isnan(self.coefficients), axis=1)


class FDataBasisDType(pandas.api.extensions.ExtensionDtype):  # type: ignore
    """DType corresponding to FDataBasis in Pandas."""

    kind = 'O'
    type = FDataBasis  # noqa: WPS125
    name = 'FDataBasis'
    na_value = pandas.NA

    _metadata = ("basis")

    def __init__(self, basis: Basis) -> None:
        self.basis = basis

    @classmethod
    def construct_array_type(cls) -> Type[FDataBasis]:  # noqa: D102
        return FDataBasis

    def _na_repr(self) -> FDataBasis:
        return FDataBasis(
            basis=self.basis,
            coefficients=((np.NaN,) * self.basis.n_basis,),
        )

    def __eq__(self, other: Any) -> bool:
        """
        Compare dtype equality.

        Rules for equality (similar to categorical):
        1) Any FData is equal to the string 'category'
        2) Any FData is equal to itself
        3) Otherwise, they are equal if the arguments are equal.
        6) Any other comparison returns False
        """
        if isinstance(other, str):
            return other == self.name
        elif other is self:
            return True

        return (
            isinstance(other, FDataBasisDType)
            and self.basis == other.basis
        )

    def __hash__(self) -> int:
        return hash(self.basis)


class _CoordinateIterator(Sequence[T]):
    """Internal class to iterate through the image coordinates.

    Dummy object. Should be change to support multidimensional objects.

    """

    def __init__(self, fdatabasis: T) -> None:
        """Create an iterator through the image coordinates."""
        self._fdatabasis = fdatabasis

    def __getitem__(self, key: Union[int, slice]) -> T:
        """Get a specific coordinate."""
        basis, coefs = self._fdatabasis.basis.coordinate_basis_and_coefs(
            self._fdatabasis.coefficients,
            key,
        )

        coord_names = self._fdatabasis.coordinate_names[key]
        if coord_names is None or isinstance(coord_names, str):
            coord_names = (coord_names,)

        return self._fdatabasis.copy(
            basis=basis,
            coefficients=coefs,
            coordinate_names=coord_names,
        )

    def __len__(self) -> int:
        """Return the number of coordinates."""
        return self._fdatabasis.dim_codomain



import unittest

import numpy as np
from scipy.integrate import cumtrapz

from skfda.datasets import make_gaussian, make_gaussian_process
from skfda.misc.covariances import Gaussian
from skfda.misc.operators import LinearDifferentialOperator
from skfda.misc.regularization import L2Regularization
from skfda.ml.regression import HistoricalLinearRegression, LinearRegression
from skfda.representation.basis import BSpline, FDataBasis, Fourier, Monomial
from skfda.representation.grid import FDataGrid


class TestScalarLinearRegression(unittest.TestCase):

    def test_regression_single_explanatory(self):

        x_basis = Monomial(n_basis=7)
        x_fd = FDataBasis(x_basis, np.identity(7))

        beta_basis = Fourier(n_basis=5)
        beta_fd = FDataBasis(beta_basis, [1, 1, 1, 1, 1])
        y = [0.9999999999999993,
             0.162381381441085,
             0.08527083481359901,
             0.08519946930844623,
             0.09532291032042489,
             0.10550022969639987,
             0.11382675064746171]

        scalar = LinearRegression(coef_basis=[beta_basis])
        scalar.fit(x_fd, y)
        np.testing.assert_allclose(scalar.coef_[0].coefficients,
                                   beta_fd.coefficients)
        np.testing.assert_allclose(scalar.intercept_,
                                   0.0, atol=1e-6)

        y_pred = scalar.predict(x_fd)
        np.testing.assert_allclose(y_pred, y)

        scalar = LinearRegression(coef_basis=[beta_basis],
                                  fit_intercept=False)
        scalar.fit(x_fd, y)
        np.testing.assert_allclose(scalar.coef_[0].coefficients,
                                   beta_fd.coefficients)
        np.testing.assert_equal(scalar.intercept_,
                                0.0)

        y_pred = scalar.predict(x_fd)
        np.testing.assert_allclose(y_pred, y)

    def test_regression_multiple_explanatory(self):
        y = [1, 2, 3, 4, 5, 6, 7]

        X = FDataBasis(Monomial(n_basis=7), np.identity(7))

        beta1 = BSpline(domain_range=(0, 1), n_basis=5)

        scalar = LinearRegression(coef_basis=[beta1])

        scalar.fit(X, y)

        np.testing.assert_allclose(scalar.intercept_.round(4),
                                   np.array([32.65]), rtol=1e-3)

        np.testing.assert_allclose(
            scalar.coef_[0].coefficients.round(4),
            np.array([[-28.6443,
                       80.3996,
                       -188.587,
                       236.5832,
                       -481.3449]]), rtol=1e-3)

        y_pred = scalar.predict(X)
        np.testing.assert_allclose(y_pred, y, atol=0.01)

    def test_regression_mixed(self):

        multivariate = np.array([[0, 0], [2, 7], [1, 7], [3, 9],
                                 [4, 16], [2, 14], [3, 5]])

        X = [multivariate,
             FDataBasis(Monomial(n_basis=3), [[1, 0, 0], [0, 1, 0], [0, 0, 1],
                                              [1, 0, 1], [1, 0, 0], [0, 1, 0],
                                              [0, 0, 1]])]

        # y = 2 + sum([3, 1] * array) + int(3 * function)
        intercept = 2
        coefs_multivariate = np.array([3, 1])
        coefs_functions = FDataBasis(
            Monomial(n_basis=3), [[3, 0, 0]])
        y_integral = np.array([3, 3 / 2, 1, 4, 3, 3 / 2, 1])
        y_sum = multivariate @ coefs_multivariate
        y = 2 + y_sum + y_integral

        scalar = LinearRegression()
        scalar.fit(X, y)

        np.testing.assert_allclose(scalar.intercept_,
                                   intercept, atol=0.01)

        np.testing.assert_allclose(
            scalar.coef_[0],
            coefs_multivariate, atol=0.01)

        np.testing.assert_allclose(
            scalar.coef_[1].coefficients,
            coefs_functions.coefficients, atol=0.01)

        y_pred = scalar.predict(X)
        np.testing.assert_allclose(y_pred, y, atol=0.01)

    def test_regression_mixed_regularization(self):

        multivariate = np.array([[0, 0], [2, 7], [1, 7], [3, 9],
                                 [4, 16], [2, 14], [3, 5]])

        X = [multivariate,
             FDataBasis(Monomial(n_basis=3), [[1, 0, 0], [0, 1, 0], [0, 0, 1],
                                              [1, 0, 1], [1, 0, 0], [0, 1, 0],
                                              [0, 0, 1]])]

        # y = 2 + sum([3, 1] * array) + int(3 * function)
        intercept = 2
        coefs_multivariate = np.array([3, 1])
        y_integral = np.array([3, 3 / 2, 1, 4, 3, 3 / 2, 1])
        y_sum = multivariate @ coefs_multivariate
        y = 2 + y_sum + y_integral

        scalar = LinearRegression(
            regularization=[L2Regularization(lambda x: x),
                            L2Regularization(
                                LinearDifferentialOperator(2))])
        scalar.fit(X, y)

        np.testing.assert_allclose(scalar.intercept_,
                                   intercept, atol=0.01)

        np.testing.assert_allclose(
            scalar.coef_[0],
            [2.536739, 1.072186], atol=0.01)

        np.testing.assert_allclose(
            scalar.coef_[1].coefficients,
            [[2.125676, 2.450782, 5.808745e-4]], atol=0.01)

        y_pred = scalar.predict(X)
        np.testing.assert_allclose(
            y_pred,
            [5.349035, 16.456464, 13.361185, 23.930295,
                32.650965, 23.961766, 16.29029],
            atol=0.01)

    def test_regression_regularization(self):

        x_basis = Monomial(n_basis=7)
        x_fd = FDataBasis(x_basis, np.identity(7))

        beta_basis = Fourier(n_basis=5)
        beta_fd = FDataBasis(beta_basis, [1.0403, 0, 0, 0, 0])
        y = [1.0000684777229512,
             0.1623672257830915,
             0.08521053851548224,
             0.08514200869281137,
             0.09529138749665378,
             0.10549625973303875,
             0.11384314859153018]

        y_pred_compare = [0.890341,
                          0.370162,
                          0.196773,
                          0.110079,
                          0.058063,
                          0.023385,
                          -0.001384]

        scalar = LinearRegression(
            coef_basis=[beta_basis],
            regularization=L2Regularization(
                LinearDifferentialOperator(2)))
        scalar.fit(x_fd, y)
        np.testing.assert_allclose(scalar.coef_[0].coefficients,
                                   beta_fd.coefficients, atol=1e-3)
        np.testing.assert_allclose(scalar.intercept_,
                                   -0.15, atol=1e-4)

        y_pred = scalar.predict(x_fd)
        np.testing.assert_allclose(y_pred, y_pred_compare, atol=1e-4)

        x_basis = Monomial(n_basis=3)
        x_fd = FDataBasis(x_basis, [[1, 0, 0],
                                    [0, 1, 0],
                                    [0, 0, 1],
                                    [2, 0, 1]])

        beta_fd = FDataBasis(x_basis, [3, 2, 1])
        y = [1 + 13 / 3, 1 + 29 / 12, 1 + 17 / 10, 1 + 311 / 30]

        # Non regularized
        scalar = LinearRegression()
        scalar.fit(x_fd, y)
        np.testing.assert_allclose(scalar.coef_[0].coefficients,
                                   beta_fd.coefficients)
        np.testing.assert_allclose(scalar.intercept_,
                                   1)

        y_pred = scalar.predict(x_fd)
        np.testing.assert_allclose(y_pred, y)

        # Regularized
        beta_fd_reg = FDataBasis(x_basis, [2.812, 3.043, 0])
        y_reg = [5.333, 3.419, 2.697, 11.366]

        scalar_reg = LinearRegression(
            regularization=L2Regularization(
                LinearDifferentialOperator(2)))
        scalar_reg.fit(x_fd, y)
        np.testing.assert_allclose(scalar_reg.coef_[0].coefficients,
                                   beta_fd_reg.coefficients, atol=0.001)
        np.testing.assert_allclose(scalar_reg.intercept_,
                                   0.998, atol=0.001)

        y_pred = scalar_reg.predict(x_fd)
        np.testing.assert_allclose(y_pred, y_reg, atol=0.001)

    def test_error_X_not_FData(self):
        """Tests that at least one of the explanatory variables
        is an FData object. """

        x_fd = np.identity(7)
        y = np.zeros(7)

        scalar = LinearRegression(coef_basis=[Fourier(n_basis=5)])

        with np.testing.assert_warns(UserWarning):
            scalar.fit([x_fd], y)

    def test_error_y_is_FData(self):
        """Tests that none of the explained variables is an FData object
        """
        x_fd = FDataBasis(Monomial(n_basis=7), np.identity(7))
        y = list(FDataBasis(Monomial(n_basis=7), np.identity(7)))

        scalar = LinearRegression(coef_basis=[Fourier(n_basis=5)])

        with np.testing.assert_raises(ValueError):
            scalar.fit([x_fd], y)

    def test_error_X_beta_len_distinct(self):
        """ Test that the number of beta bases and explanatory variables
        are not different """

        x_fd = FDataBasis(Monomial(n_basis=7), np.identity(7))
        y = [1 for _ in range(7)]
        beta = Fourier(n_basis=5)

        scalar = LinearRegression(coef_basis=[beta])
        with np.testing.assert_raises(ValueError):
            scalar.fit([x_fd, x_fd], y)

        scalar = LinearRegression(coef_basis=[beta, beta])
        with np.testing.assert_raises(ValueError):
            scalar.fit([x_fd], y)

    def test_error_y_X_samples_different(self):
        """ Test that the number of response samples and explanatory samples
        are not different """

        x_fd = FDataBasis(Monomial(n_basis=7), np.identity(7))
        y = [1 for _ in range(8)]
        beta = Fourier(n_basis=5)

        scalar = LinearRegression(coef_basis=[beta])
        with np.testing.assert_raises(ValueError):
            scalar.fit([x_fd], y)

        x_fd = FDataBasis(Monomial(n_basis=8), np.identity(8))
        y = [1 for _ in range(7)]
        beta = Fourier(n_basis=5)

        scalar = LinearRegression(coef_basis=[beta])
        with np.testing.assert_raises(ValueError):
            scalar.fit([x_fd], y)

    def test_error_beta_not_basis(self):
        """ Test that all beta are Basis objects. """

        x_fd = FDataBasis(Monomial(n_basis=7), np.identity(7))
        y = [1 for _ in range(7)]
        beta = FDataBasis(Monomial(n_basis=7), np.identity(7))

        scalar = LinearRegression(coef_basis=[beta])
        with np.testing.assert_raises(TypeError):
            scalar.fit([x_fd], y)

    def test_error_weights_lenght(self):
        """ Test that the number of weights is equal to the
        number of samples """

        x_fd = FDataBasis(Monomial(n_basis=7), np.identity(7))
        y = [1 for _ in range(7)]
        weights = [1 for _ in range(8)]
        beta = Monomial(n_basis=7)

        scalar = LinearRegression(coef_basis=[beta])
        with np.testing.assert_raises(ValueError):
            scalar.fit([x_fd], y, weights)

    def test_error_weights_negative(self):
        """ Test that none of the weights are negative. """

        x_fd = FDataBasis(Monomial(n_basis=7), np.identity(7))
        y = [1 for _ in range(7)]
        weights = [-1 for _ in range(7)]
        beta = Monomial(n_basis=7)

        scalar = LinearRegression(coef_basis=[beta])
        with np.testing.assert_raises(ValueError):
            scalar.fit([x_fd], y, weights)


class TestHistoricalLinearRegression(unittest.TestCase):
    """Tests for historical linear regression."""

    def setUp(self) -> None:
        """Generate data according to the model."""
        self.random = np.random.RandomState(1)

        self.n_samples = 50
        self.n_features = 20
        self.intercept = make_gaussian_process(
            n_samples=1,
            n_features=self.n_features,
            cov=Gaussian(length_scale=0.4),
            random_state=self.random,
        )

        self.X = make_gaussian_process(
            n_samples=self.n_samples,
            n_features=self.n_features,
            cov=Gaussian(length_scale=0.4),
            random_state=self.random,
        )

        self.coefficients = make_gaussian(
            n_samples=1,
            grid_points=[np.linspace(0, 1, self.n_features)] * 2,
            cov=Gaussian(length_scale=1),
            random_state=self.random,
        )

        self.X2 = make_gaussian_process(
            n_samples=self.n_samples,
            n_features=self.n_features,
            cov=Gaussian(length_scale=0.4),
            random_state=self.random,
        )

        self.coefficients2 = make_gaussian(
            n_samples=1,
            grid_points=[np.linspace(0, 1, self.n_features)] * 2,
            cov=Gaussian(length_scale=1),
            random_state=self.random,
        )

        self.create_model()
        self.create_vectorial_model()

    def create_model_no_intercept(
        self,
        X: FDataGrid,
        coefficients: FDataGrid,
    ) -> FDataGrid:
        """Create a functional response according to historical model."""
        integral_body = (
            X.data_matrix[..., 0, np.newaxis]
            * coefficients.data_matrix[..., 0]
        )
        integral_matrix = cumtrapz(
            integral_body,
            x=X.grid_points[0],
            initial=0,
            axis=1,
        )
        integral = np.diagonal(integral_matrix, axis1=1, axis2=2)
        return X.copy(data_matrix=integral)

    def create_model(self) -> None:
        """Create a functional response according to historical model."""
        model_no_intercept = self.create_model_no_intercept(
            X=self.X,
            coefficients=self.coefficients,
        )
        self.y = model_no_intercept + self.intercept

    def create_vectorial_model(self) -> None:
        """Create a functional response according to historical model."""
        model_no_intercept = self.create_model_no_intercept(
            X=self.X,
            coefficients=self.coefficients,
        )
        model_no_intercept2 = self.create_model_no_intercept(
            X=self.X2,
            coefficients=self.coefficients2,
        )
        self.y2 = model_no_intercept + model_no_intercept2 + self.intercept

    def test_historical(self) -> None:
        """Test historical regression with data following the model."""
        regression = HistoricalLinearRegression(n_intervals=6)
        fit_predict_result = regression.fit_predict(self.X, self.y)
        predict_result = regression.predict(self.X)

        np.testing.assert_allclose(
            predict_result.data_matrix,
            fit_predict_result.data_matrix,
        )

        np.testing.assert_allclose(
            predict_result.data_matrix,
            self.y.data_matrix,
            rtol=1e-1,
        )

        np.testing.assert_allclose(
            regression.intercept_.data_matrix,
            self.intercept.data_matrix,
            rtol=1e-3,
        )

        np.testing.assert_allclose(
            regression.coef_.data_matrix[0, ..., 0],
            np.triu(self.coefficients.data_matrix[0, ..., 0]),
            atol=0.3,
            rtol=0,
        )

    def test_historical_vectorial(self) -> None:
        """Test historical regression with data following the vector model."""
        X = self.X.concatenate(self.X2, as_coordinates=True)

        regression = HistoricalLinearRegression(n_intervals=10)
        fit_predict_result = regression.fit_predict(X, self.y2)
        predict_result = regression.predict(X)

        np.testing.assert_allclose(
            predict_result.data_matrix,
            fit_predict_result.data_matrix,
        )

        np.testing.assert_allclose(
            predict_result.data_matrix,
            self.y2.data_matrix,
            atol=1e-1,
            rtol=0,
        )

        np.testing.assert_allclose(
            regression.intercept_.data_matrix,
            self.intercept.data_matrix,
            rtol=1e-2,
        )

        # Coefficient matrix not tested as it is probably
        # an ill-posed problem


if __name__ == '__main__':
    print()
    unittest.main()



"""Typing for norms and metrics."""
import enum
from abc import abstractmethod
from builtins import isinstance
from typing import Any, TypeVar, Union, overload

from typing_extensions import Final, Literal, Protocol

from ...representation._typing import NDArrayFloat, Vector

VectorType = TypeVar("VectorType", contravariant=True, bound=Vector)
MetricElementType = TypeVar("MetricElementType", contravariant=True)


class _MetricSingletons(enum.Enum):
    PRECOMPUTED = "precomputed"


PRECOMPUTED: Final = _MetricSingletons.PRECOMPUTED

_PrecomputedTypes = Literal[
    _MetricSingletons.PRECOMPUTED,
    "precomputed",
]


class Norm(Protocol[VectorType]):
    """Protocol for a norm of a vector."""

    @abstractmethod
    def __call__(self, __vector: VectorType) -> NDArrayFloat:  # noqa: WPS112
        """Compute the norm of a vector."""


class Metric(Protocol[MetricElementType]):
    """Protocol for a metric between two elements of a metric space."""

    @abstractmethod
    def __call__(
        self,
        __e1: MetricElementType,  # noqa: WPS112
        __e2: MetricElementType,  # noqa: WPS112
    ) -> NDArrayFloat:
        """Compute the metric between two vectors."""


_NonStringMetric = TypeVar(
    "_NonStringMetric",
    bound=Union[
        Metric[Any],
        _MetricSingletons,
    ],
)


@overload
def _parse_metric(
    metric: str,
) -> _MetricSingletons:
    pass


@overload
def _parse_metric(
    metric: _NonStringMetric,
) -> _NonStringMetric:
    pass


def _parse_metric(
    metric: Union[Metric[Any], _MetricSingletons, str],
) -> Union[Metric[Any], _MetricSingletons]:

    return _MetricSingletons(metric) if isinstance(metric, str) else metric



import operator

import numpy as np
import pandas
import pytest
from pandas import Series
from pandas.tests.extension import base

import skfda
from skfda.representation.basis import BSpline, Fourier, Monomial


##############################################################################
# Fixtures
##############################################################################
@pytest.fixture(params=[Monomial(n_basis=5), Fourier(n_basis=5),
                        BSpline(n_basis=5)])
def basis(request):
    """A fixture providing the ExtensionDtype to validate."""

    return request.param


@pytest.fixture
def dtype(basis):
    """A fixture providing the ExtensionDtype to validate."""

    return skfda.representation.basis.FDataBasisDType(basis=basis)


@pytest.fixture
def data(basis):
    """
    Length-100 array for this type.
    * data[0] and data[1] should both be non missing
    * data[0] and data[1] should not be equal
    """

    coef_matrix = np.arange(100 * basis.n_basis).reshape(100, basis.n_basis)

    return skfda.FDataBasis(basis=basis, coefficients=coef_matrix)


@pytest.fixture
def data_for_twos():
    """Length-100 array in which all the elements are two."""
    raise NotImplementedError


@pytest.fixture
def data_missing(basis):
    """Length-2 array with [NA, Valid]"""

    coef_matrix = np.arange(
        2 * basis.n_basis, dtype=np.float_).reshape(2, basis.n_basis)
    coef_matrix[0, :] = np.NaN

    return skfda.FDataBasis(basis=basis, coefficients=coef_matrix)


@pytest.fixture(params=["data", "data_missing"])
def all_data(request, data, data_missing):
    """Parametrized fixture giving 'data' and 'data_missing'"""
    if request.param == "data":
        return data
    elif request.param == "data_missing":
        return data_missing


@pytest.fixture
def data_repeated(data):
    """
    Generate many datasets.
    Parameters
    ----------
    data : fixture implementing `data`
    Returns
    -------
    Callable[[int], Generator]:
        A callable that takes a `count` argument and
        returns a generator yielding `count` datasets.
    """

    def gen(count):
        for _ in range(count):
            yield data

    return gen


@pytest.fixture
def data_for_sorting():
    """
    Length-3 array with a known sort order.
    This should be three items [B, C, A] with
    A < B < C
    """
    raise NotImplementedError


@pytest.fixture
def data_missing_for_sorting():
    """
    Length-3 array with a known sort order.
    This should be three items [B, NA, A] with
    A < B and NA missing.
    """
    raise NotImplementedError


@pytest.fixture
def na_cmp():
    """
    Binary operator for comparing NA values.
    Should return a function of two arguments that returns
    True if both arguments are (scalar) NA for your type.
    By default, uses ``operator.is_``
    """
    def isna(x, y):
        return ((x is pandas.NA or all(x.isna()))
                and (y is pandas.NA or all(y.isna())))

    return isna


@pytest.fixture
def na_value():
    """The scalar missing value for this type. Default 'None'"""
    return pandas.NA


@pytest.fixture
def data_for_grouping():
    """
    Data for factorization, grouping, and unique tests.
    Expected to be like [B, B, NA, NA, A, A, B, C]
    Where A < B < C and NA is missing
    """
    raise NotImplementedError


@pytest.fixture(params=[True, False])
def box_in_series(request):
    """Whether to box the data in a Series"""
    return request.param


@pytest.fixture(
    params=[
        lambda x: 1,
        lambda x: [1] * len(x),
        lambda x: Series([1] * len(x)),
        lambda x: x,
    ],
    ids=["scalar", "list", "series", "object"],
)
def groupby_apply_op(request):
    """
    Functions to test groupby.apply().
    """
    return request.param


@pytest.fixture(params=[True, False])
def as_frame(request):
    """
    Boolean fixture to support Series and Series.to_frame() comparison testing.
    """
    return request.param


@pytest.fixture(params=[True, False])
def as_series(request):
    """
    Boolean fixture to support arr and Series(arr) comparison testing.
    """
    return request.param


@pytest.fixture(params=[True, False])
def use_numpy(request):
    """
    Boolean fixture to support comparison testing of ExtensionDtype array
    and numpy array.
    """
    return request.param


@pytest.fixture(params=["ffill", "bfill"])
def fillna_method(request):
    """
    Parametrized fixture giving method parameters 'ffill' and 'bfill' for
    Series.fillna(method=<method>) testing.
    """
    return request.param


@pytest.fixture(params=[True, False])
def as_array(request):
    """
    Boolean fixture to support ExtensionDtype _from_sequence method testing.
    """
    return request.param


_all_arithmetic_operators = [
    "__add__",
    "__radd__",
    "__sub__",
    "__rsub__",
    #     "__mul__",
    #     "__rmul__",
    #     "__floordiv__",
    #     "__rfloordiv__",
    #     "__truediv__",
    #     "__rtruediv__",
    #     "__pow__",
    #     "__rpow__",
    #     "__mod__",
    #     "__rmod__",
]


@pytest.fixture(params=_all_arithmetic_operators)
def all_arithmetic_operators(request):
    """
    Fixture for dunder names for common arithmetic operations.
    """
    return request.param


@pytest.fixture(params=["__eq__", "__ne__",
                        # "__le__", "__lt__", "__ge__", "__gt__"
                        ])
def all_compare_operators(request):
    """
    Fixture for dunder names for common compare operations
    """
    return request.param


_all_numeric_reductions = [
    "sum",
    #     "max",
    #     "min",
    "mean",
    #     "prod",
    #     "std",
    #     "var",
    #     "median",
    #     "kurt",
    #     "skew",
]


@pytest.fixture(params=_all_numeric_reductions)
def all_numeric_reductions(request):
    """
    Fixture for numeric reduction names.
    """
    return request.param

##############################################################################
# Tests
##############################################################################


class TestCasting(base.BaseCastingTests):

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_astype_str(self):
        pass

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_astype_string(self):
        pass


class TestConstructors(base.BaseConstructorsTests):

    # Does not support scalars which are also ExtensionArrays
    @pytest.mark.skip(reason="Unsupported")
    def test_series_constructor_scalar_with_index(self):
        pass

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_from_dtype(self):
        pass


class TestDtype(base.BaseDtypeTests):

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_construct_from_string_own_name(self):
        pass

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_is_dtype_from_name(self):
        pass

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_eq_with_str(self):
        pass

    # Tries to construct dtype from string
    @pytest.mark.skip(reason="Unsupported")
    def test_construct_from_string(self, dtype):
        pass


class TestGetitem(base.BaseGetitemTests):
    pass


class TestInterface(base.BaseInterfaceTests):

    # Does not support scalars which are also array_like
    @pytest.mark.skip(reason="Unsupported")
    def test_array_interface(self):
        pass

    # We do not implement setitem
    @pytest.mark.skip(reason="Unsupported")
    def test_copy(self, dtype):
        pass

    # We do not implement setitem
    @pytest.mark.skip(reason="Unsupported")
    def test_view(self, dtype):
        pass

    # Pending https://github.com/pandas-dev/pandas/issues/38812 resolution
    @pytest.mark.skip(reason="Bugged")
    def test_contains(self, data, data_missing):
        pass


class TestArithmeticOps(base.BaseArithmeticOpsTests):

    series_scalar_exc = None

    # Bug introduced by https://github.com/pandas-dev/pandas/pull/37132
    @pytest.mark.skip(reason="Unsupported")
    def test_arith_frame_with_scalar(self, data, all_arithmetic_operators):
        pass

    # FDatabasis does not implement division by non constant
    @pytest.mark.skip(reason="Unsupported")
    def test_divmod_series_array(self, dtype):
        pass

    # Does not convert properly a list of FData to a FData
    @pytest.mark.skip(reason="Unsupported")
    def test_arith_series_with_array(self, dtype):
        pass

    # Does not error on operations
    @pytest.mark.skip(reason="Unsupported")
    def test_error(self, dtype):
        pass


class TestComparisonOps(base.BaseComparisonOpsTests):

    # Cannot be compared with 0
    @pytest.mark.skip(reason="Unsupported")
    def test_compare_scalar(self, data, all_compare_operators):
        pass

    # Not sure how to pass it. Should it be reimplemented?
    @pytest.mark.skip(reason="Unsupported")
    def test_compare_array(self, data, all_compare_operators):
        pass


class TestNumericReduce(base.BaseNumericReduceTests):

    def check_reduce(self, s, op_name, skipna):
        result = getattr(s, op_name)(skipna=skipna)
        assert result.n_samples == 1



from __future__ import annotations

import abc
from typing import Callable, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import sklearn.gaussian_process.kernels as sklearn_kern
from matplotlib.figure import Figure
from scipy.special import gamma, kv

from ..representation._typing import ArrayLike, NDArrayFloat


def _squared_norms(x: NDArrayFloat, y: NDArrayFloat) -> NDArrayFloat:
    return ((x[np.newaxis, :, :] - y[:, np.newaxis, :]) ** 2).sum(2)


CovarianceLike = Union[
    float,
    NDArrayFloat,
    Callable[[ArrayLike, ArrayLike], NDArrayFloat],
]


def _transform_to_2d(t: ArrayLike) -> NDArrayFloat:
    """Transform 1d arrays in column vectors."""
    t = np.asarray(t)

    dim = len(t.shape)
    assert dim <= 2

    if dim < 2:
        t = np.atleast_2d(t).T

    return t


def _execute_covariance(
    covariance: CovarianceLike,
    x: ArrayLike,
    y: ArrayLike,
) -> NDArrayFloat:
    """Execute a covariance function."""
    x = _transform_to_2d(x)
    y = _transform_to_2d(y)

    if isinstance(covariance, (int, float)):
        return np.array(covariance)
    else:
        if callable(covariance):
            result = covariance(x, y)
        elif isinstance(covariance, np.ndarray):
            result = covariance
        else:
            # GPy kernel
            result = covariance.K(x, y)

        assert result.shape[0] == len(x)
        assert result.shape[1] == len(y)
        return result


class Covariance(abc.ABC):
    """Abstract class for covariance functions."""

    _parameters_str: Sequence[Tuple[str, str]]
    _latex_formula: str

    @abc.abstractmethod
    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        pass

    def heatmap(self, limits: Tuple[float, float] = (-1, 1)) -> Figure:
        """Return a heatmap plot of the covariance function."""
        from ..exploratory.visualization._utils import _create_figure

        x = np.linspace(*limits, 1000)

        cov_matrix = self(x, x)

        fig = _create_figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.imshow(
            cov_matrix,
            extent=[limits[0], limits[1], limits[1], limits[0]],
        )
        ax.set_title(f"Covariance function in [{limits[0]}, {limits[1]}]")

        return fig

    def _sample_trajectories_plot(self) -> Figure:
        from ..datasets import make_gaussian_process

        fd = make_gaussian_process(
            start=-1,
            n_samples=10,
            cov=self,
            random_state=0,
        )
        fig = fd.plot()
        fig.axes[0].set_title("Sample trajectories")
        return fig

    def __repr__(self) -> str:

        params_str = ', '.join(
            f'{n}={getattr(self, n)}' for n, _ in self._parameters_str
        )

        return (
            f"{self.__module__}.{type(self).__qualname__}("
            f"{params_str}"
            f")"
        )

    def _latex_content(self) -> str:
        params_str = ''.join(
            fr'{l} &= {getattr(self, n)} \\' for n, l in self._parameters_str
        )

        return (
            fr"{self._latex_formula} \\"
            r"\text{where:}"
            r"\begin{aligned}"
            fr"\qquad{params_str}"
            r"\end{aligned}"
        )

    def _repr_latex_(self) -> str:
        return fr"\(\displaystyle {self._latex_content()}\)"

    def _repr_html_(self) -> str:
        from ..exploratory.visualization._utils import _figure_to_svg

        fig = self.heatmap()
        heatmap = _figure_to_svg(fig)
        plt.close(fig)

        fig = self._sample_trajectories_plot()
        sample_trajectories = _figure_to_svg(fig)
        plt.close(fig)

        row_style = ''

        def column_style(percent: float, margin_top: str = "0") -> str:
            return (
                f'style="display: inline-block; '
                f'margin:0; '
                f'margin-top: {margin_top}; '
                f'width:{percent}%; '
                f'height:auto;'
                f'vertical-align: middle"'
            )

        return fr"""
        <div {row_style}>
            <div {column_style(100, margin_top='25px')}>
            \[{self._latex_content()}\]
            </div>
        </div>
        <div {row_style}>
            <div {column_style(48)}>
            {sample_trajectories}
            </div>
            <div {column_style(48)}>
            {heatmap}
            </div>
        </div>
        """

    def to_sklearn(self) -> sklearn_kern.Kernel:
        """Convert it to a sklearn kernel, if there is one."""
        raise NotImplementedError(
            f"{type(self).__name__} covariance not "
            f"implemented in scikit-learn",
        )


class Brownian(Covariance):
    r"""
    Brownian covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 \frac{|x - \mathcal{O}| + |x' - \mathcal{O}|
        - |x - x'|}{2}

    where :math:`\sigma^2` is the variance at distance 1 from
    :math:`\mathcal{O}` and :math:`\mathcal{O}` is the origin point.
    If :math:`\mathcal{O} = 0` (the default) and we only
    consider positive values, the formula can be simplified as

    .. math::
        K(x, y) = \sigma^2 \min(x, y).

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Brownian
        import matplotlib.pyplot as plt

        Brownian().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Brownian
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Brownian(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Brownian

        Brownian()

    """
    _latex_formula = (
        r"K(x, x') = \sigma^2 \frac{|x - \mathcal{O}| + "
        r"|x' - \mathcal{O}| - |x - x'|}{2}"
    )

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("origin", r"\mathcal{O}"),
    ]

    def __init__(self, *, variance: float = 1, origin: float = 0) -> None:
        self.variance = variance
        self.origin = origin

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x) - self.origin
        y = _transform_to_2d(y) - self.origin

        sum_norms = np.add.outer(
            np.linalg.norm(x, axis=-1),
            np.linalg.norm(y, axis=-1),
        )
        norm_sub = np.linalg.norm(
            x[:, np.newaxis, :] - y[np.newaxis, :, :],
            axis=-1,
        )

        return self.variance * (sum_norms - norm_sub) / 2


class Linear(Covariance):
    r"""
    Linear covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 (x^T x' + c)

    where :math:`\sigma^2` is the scale of the variance and
    :math:`c` is the intercept.

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Linear
        import matplotlib.pyplot as plt

        Linear().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Linear
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Linear(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Linear

        Linear()

    """

    _latex_formula = r"K(x, x') = \sigma^2 (x^T x' + c)"

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("intercept", "c"),
    ]

    def __init__(self, *, variance: float = 1, intercept: float = 0) -> None:
        self.variance = variance
        self.intercept = intercept

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        y = _transform_to_2d(y)

        return self.variance * (x @ y.T + self.intercept)

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return (
            self.variance
            * (sklearn_kern.DotProduct(0) + self.intercept)
        )


class Polynomial(Covariance):
    r"""
    Polynomial covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 (\alpha x^T x' + c)^d

    where :math:`\sigma^2` is the scale of the variance,
    :math:`\alpha` is the slope, :math:`d` the degree of the
    polynomial and :math:`c` is the intercept.

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Polynomial
        import matplotlib.pyplot as plt

        Polynomial().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Polynomial
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Polynomial(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Polynomial

        Polynomial()

    """

    _latex_formula = r"K(x, x') = \sigma^2 (\alpha x^T x' + c)^d"

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("intercept", "c"),
        ("slope", r"\alpha"),
        ("degree", "d"),
    ]

    def __init__(
        self,
        *,
        variance: float = 1,
        intercept: float = 0,
        slope: float = 1,
        degree: float = 2,
    ) -> None:
        self.variance = variance
        self.intercept = intercept
        self.slope = slope
        self.degree = degree

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        y = _transform_to_2d(y)

        return (
            self.variance
            * (self.slope * x @ y.T + self.intercept) ** self.degree
        )

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return (
            self.variance
            * (self.slope * sklearn_kern.DotProduct(0) + self.intercept)
            ** self.degree
        )


class Gaussian(Covariance):
    r"""
    Gaussian covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 \exp\left(-\frac{||x - x'||^2}{2l^2}\right)

    where :math:`\sigma^2` is the variance and :math:`l` is the length scale.

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Gaussian
        import matplotlib.pyplot as plt

        Gaussian().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Gaussian
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Gaussian(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Gaussian

        Gaussian()

    """

    _latex_formula = (
        r"K(x, x') = \sigma^2 \exp\left(-\frac{\|x - x'\|^2}{2l^2}"
        r"\right)"
    )

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("length_scale", "l"),
    ]

    def __init__(self, *, variance: float = 1, length_scale: float = 1):
        self.variance = variance
        self.length_scale = length_scale

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        y = _transform_to_2d(y)

        x_y = _squared_norms(x, y)

        return self.variance * np.exp(-x_y / (2 * self.length_scale ** 2))

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return (
            self.variance * sklearn_kern.RBF(length_scale=self.length_scale)
        )


class Exponential(Covariance):
    r"""
    Exponential covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 \exp\left(-\frac{\|x - x'\|}{l}\right)

    where :math:`\sigma^2` is the variance and :math:`l` is the length scale.

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Exponential
        import matplotlib.pyplot as plt

        Exponential().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Exponential
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Exponential(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Exponential

        Exponential()

    """

    _latex_formula = (
        r"K(x, x') = \sigma^2 \exp\left(-\frac{||x - x'||}{l}"
        r"\right)"
    )

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("length_scale", "l"),
    ]

    def __init__(
        self,
        *,
        variance: float = 1,
        length_scale: float = 1,
    ) -> None:
        self.variance = variance
        self.length_scale = length_scale

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        y = _transform_to_2d(y)

        x_y = _squared_norms(x, y)
        return self.variance * np.exp(-np.sqrt(x_y) / (self.length_scale))

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return (
            self.variance
            * sklearn_kern.Matern(length_scale=self.length_scale, nu=0.5)
        )


class WhiteNoise(Covariance):
    r"""
    Gaussian covariance function.

    The covariance function is

    .. math::
        K(x, x')= \begin{cases}
                    \sigma^2, \quad x = x' \\
                    0, \quad x \neq x'\\
                  \end{cases}

    where :math:`\sigma^2` is the variance.

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import WhiteNoise
        import matplotlib.pyplot as plt

        WhiteNoise().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import WhiteNoise
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=WhiteNoise(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import WhiteNoise

        WhiteNoise()

    """

    _latex_formula = (
        r"K(x, x')= \begin{cases} \sigma^2, \quad x = x' \\"
        r"0, \quad x \neq x'\\ \end{cases}"
    )

    _parameters_str = [("variance", r"\sigma^2")]

    def __init__(self, *, variance: float = 1):
        self.variance = variance

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        return self.variance * np.eye(x.shape[0])

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return sklearn_kern.WhiteKernel(noise_level=self.variance)


class Matern(Covariance):
    r"""
    Matérn covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 \frac{2^{1-\nu}}{\Gamma(\nu)}
        \left( \frac{\sqrt{2\nu}|x - x'|}{l} \right)^{\nu}
        K_{\nu}\left( \frac{\sqrt{2\nu}|x - x'|}{l} \right)

    where :math:`\sigma^2` is the variance, :math:`l` is the length scale
    and :math:`\nu` controls the smoothness of the related Gaussian process.
    The trajectories of a Gaussian process with Matérn covariance is 
    :math:`\lceil \nu \rceil - 1` times differentiable.


    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Matern
        import matplotlib.pyplot as plt

        Matern().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Matern
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Matern(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Matern

        Matern()

    """
    _latex_formula = (
        r"K(x, x') = \sigma^2 \frac{2^{1-\nu}}{\Gamma(\nu)}"
        r"\left( \frac{\sqrt{2\nu}|x - x'|}{l} \right)^{\nu}"
        r"K_{\nu}\left( \frac{\sqrt{2\nu}|x - x'|}{l} \right)"
    )

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("length_scale", "l"),
        ("nu", r"\nu"),
    ]

    def __init__(
        self,
        *,
        variance: float = 1,
        length_scale: float = 1,
        nu: float = 1.5,
    ) -> None:
        self.variance = variance
        self.length_scale = length_scale
        self.nu = nu

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        y = _transform_to_2d(y)

        x_y_squared = _squared_norms(x, y)
        x_y = np.sqrt(x_y_squared)

        p = self.nu - 0.5
        if p.is_integer():
            # Formula for half-integers
            p = int(p)
            body = np.sqrt(2 * p + 1) * x_y / self.length_scale
            exponential = np.exp(-body)
            power_list = np.full(shape=(p,) + body.shape, fill_value=2 * body)
            power_list = np.cumprod(power_list, axis=0)
            power_list = np.concatenate(
                (power_list[::-1], [np.ones_like(body)]),
            )
            power_list = np.moveaxis(power_list, 0, -1)
            numerator = np.cumprod(np.arange(p, 0, -1))
            numerator = np.concatenate(([1], numerator))
            denom1 = np.cumprod(np.arange(2 * p, p, -1))
            denom1 = np.concatenate((denom1[::-1], [1]))
            denom2 = np.cumprod(np.arange(1, p + 1))
            denom2 = np.concatenate(([1], denom2))

            sum_terms = power_list * numerator / (denom1 * denom2)
            return self.variance * exponential * np.sum(sum_terms, axis=-1)
        elif self.nu == np.inf:
            return self.variance * np.exp(-x_y_squared / (2 * self.length_scale ** 2))
        else:
            # General formula
            scaling = 2**(1 - self.nu) / gamma(self.nu)
            body = np.sqrt(2 * self.nu) * x_y / self.length_scale
            power = body**self.nu
            bessel = kv(self.nu, body)

            with np.errstate(invalid='ignore'):
                eval_cov = self.variance * scaling * power * bessel

            # Values with nan are where the distance is 0
            return np.nan_to_num(eval_cov, nan=self.variance)

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return (
            self.variance
            * sklearn_kern.Matern(length_scale=self.length_scale, nu=self.nu)
        )



from __future__ import annotations

import math
from typing import Tuple

import numpy as np

from ...representation import FDataGrid
from ...representation._typing import NDArrayBool, NDArrayFloat, NDArrayInt


def compute_region(
    fdatagrid: FDataGrid,
    indices_descending_depth: NDArrayInt,
    prob: float,
) -> FDataGrid:
    """Compute central region of a given quantile."""
    indices_samples = indices_descending_depth[
        :math.ceil(fdatagrid.n_samples * prob)
    ]
    return fdatagrid[indices_samples]


def compute_envelope(region: FDataGrid) -> Tuple[NDArrayFloat, NDArrayFloat]:
    """Compute curves comprising a region."""
    max_envelope = np.max(region.data_matrix, axis=0)
    min_envelope = np.min(region.data_matrix, axis=0)

    return min_envelope, max_envelope


def predict_outliers(
    fdatagrid: FDataGrid,
    non_outlying_threshold: Tuple[NDArrayFloat, NDArrayFloat],
) -> NDArrayBool:
    """
    Predict outliers given a threshold.

    A functional datum is considered an outlier if it has ANY point
    in ANY dimension outside the envelope for inliers.

    """
    min_threshold, max_threshold = non_outlying_threshold

    or_axes = tuple(i for i in range(1, fdatagrid.data_matrix.ndim))

    below_outliers = np.any(
        fdatagrid.data_matrix < min_threshold,
        axis=or_axes,
    )
    above_outliers = np.any(
        fdatagrid.data_matrix > max_threshold,
        axis=or_axes,
    )

    return below_outliers | above_outliers


def non_outlying_threshold(
    central_envelope: Tuple[NDArrayFloat, NDArrayFloat],
    factor: float,
) -> Tuple[NDArrayFloat, NDArrayFloat]:
    """Compute a non outlying threshold."""
    iqr = central_envelope[1] - central_envelope[0]
    non_outlying_threshold_max = central_envelope[1] + iqr * factor
    non_outlying_threshold_min = central_envelope[0] - iqr * factor
    return (
        non_outlying_threshold_min,
        non_outlying_threshold_max,
    )



import skfda
from skfda.datasets import make_gaussian_process
from skfda.preprocessing.dim_reduction import variable_selection as vs
import unittest

import numpy as np


class TestRMH(unittest.TestCase):

    def test_rmh(self):
        n_samples = 10000
        n_features = 100

        def mean_1(t):
            return (np.abs(t - 0.25)
                    - 2 * np.abs(t - 0.5)
                    + np.abs(t - 0.75))

        X_0 = make_gaussian_process(n_samples=n_samples // 2,
                                    n_features=n_features,
                                    random_state=0)
        X_1 = make_gaussian_process(n_samples=n_samples // 2,
                                    n_features=n_features,
                                    mean=mean_1,
                                    random_state=1)
        X = skfda.concatenate((X_0, X_1))

        y = np.zeros(n_samples)
        y[n_samples // 2:] = 1

        correction = vs.recursive_maxima_hunting.GaussianSampleCorrection()
        stopping_condition = vs.recursive_maxima_hunting.ScoreThresholdStop(
            threshold=0.05)

        rmh = vs.RecursiveMaximaHunting(
            correction=correction,
            stopping_condition=stopping_condition)
        _ = rmh.fit(X, y)
        point_mask = rmh.get_support()
        points = X.grid_points[0][point_mask]
        np.testing.assert_allclose(points, [0.25, 0.5, 0.75], rtol=1e-1)


if __name__ == '__main__':
    unittest.main()



"""Elastic metrics."""

from typing import Any, Optional, Tuple, TypeVar, Union

import numpy as np
import scipy.integrate
from typing_extensions import Final

from ..._utils import normalize_scale, normalize_warping
from ...representation import FData, FDataGrid
from ...representation._typing import NDArrayFloat
from ..operators import SRSF
from ._lp_distances import l2_distance
from ._utils import PairwiseMetric, _cast_to_grid, pairwise_metric_optimization

T = TypeVar("T", bound=FData)


def _transformation_for_fisher_rao(
    fdata1: T,
    fdata2: T,
    *,
    eval_points: Optional[NDArrayFloat] = None,
    _check: bool = True,
) -> Tuple[FDataGrid, FDataGrid]:
    fdata1, fdata2 = _cast_to_grid(
        fdata1,
        fdata2,
        eval_points=eval_points,
        _check=_check,
    )

    # Both should have the same grid points
    eval_points_normalized = normalize_scale(fdata1.grid_points[0])

    # Calculate the corresponding srsf and normalize to (0,1)
    fdata1 = fdata1.copy(
        grid_points=eval_points_normalized,
        domain_range=(0, 1),
    )
    fdata2 = fdata2.copy(
        grid_points=eval_points_normalized,
        domain_range=(0, 1),
    )

    srsf = SRSF(initial_value=0)
    fdata1_srsf = srsf.fit_transform(fdata1)
    fdata2_srsf = srsf.transform(fdata2)

    return fdata1_srsf, fdata2_srsf


class FisherRaoDistance():
    r"""
    Compute the Fisher-Rao distance between two functional objects.

    Let :math:`f_i` and :math:`f_j` be two functional observations, and let
    :math:`q_i` and :math:`q_j` be the corresponding SRSF
    (see :class:`SRSF`), the Fisher-Rao distance is defined as

    .. math::
        d_{FR}(f_i, f_j) = \| q_i - q_j \|_2 =
        \left ( \int_0^1 sgn(\dot{f_i}(t))\sqrt{|\dot{f_i}(t)|} -
        sgn(\dot{f_j}(t))\sqrt{|\dot{f_j}(t)|} dt \right )^{\frac{1}{2}}

    If the observations are distributions of random variables the distance will
    match with the usual Fisher-Rao distance in non-parametric form for
    probability distributions :footcite:`srivastava++_2011_ficher-rao`.

    If the observations are defined in a :term:`domain` different than (0,1)
    their domains are normalized to this interval with an affine
    transformation.

    Args:
        fdata1: First FData object.
        fdata2: Second FData object.
        eval_points: Array with points of evaluation.

    Returns:
        Fisher rao distance.

    Raises:
        ValueError: If the objects are not unidimensional.

    References:
        .. footbibliography::

    """

    def __call__(
        self,
        fdata1: T,
        fdata2: T,
        *,
        eval_points: Optional[NDArrayFloat] = None,
        _check: bool = True,
    ) -> NDArrayFloat:
        """Compute the distance."""
        # Return the L2 distance of the SRSF
        return l2_distance(*_transformation_for_fisher_rao(
            fdata1,
            fdata2,
            eval_points=eval_points,
            _check=_check,
        ))

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}()"
        )


fisher_rao_distance: Final = FisherRaoDistance()


@pairwise_metric_optimization.register
def _pairwise_metric_optimization_fisher_rao(
    metric: FisherRaoDistance,
    elem1: T,
    elem2: Optional[T],
) -> NDArrayFloat:

    new_elem2 = elem1.copy() if elem2 is None else elem2

    new_elem1, new_elem2 = _transformation_for_fisher_rao(
        elem1,
        new_elem2,
    )

    pairwise = PairwiseMetric(l2_distance)

    return pairwise(new_elem1, None if elem2 is None else new_elem2)


def fisher_rao_amplitude_distance(
    fdata1: T,
    fdata2: T,
    *,
    lam: float = 0,
    eval_points: Optional[NDArrayFloat] = None,
    _check: bool = True,
    **kwargs: Any,
) -> NDArrayFloat:
    r"""
    Compute the Fisher-Rao amplitude distance between two functional objects.

    Let :math:`f_i` and :math:`f_j` be two functional observations, and let
    :math:`q_i` and :math:`q_j` be the corresponding SRSF
    (see :class:`SRSF`), the Fisher-Rao amplitude distance is defined as

    .. math::
        d_{A}(f_i, f_j)=min_{\gamma \in \Gamma}d_{FR}(f_i \circ \gamma,f_j)

    A penalty term could be added to restrict the ammount of elasticity in the
    alignment used.

    .. math::
        d_{\lambda}^2(f_i, f_j) =min_{\gamma \in \Gamma} \{
        d_{FR}^2(f_i \circ \gamma, f_j) + \lambda \mathcal{R}(\gamma) \}


    Where :math:`d_{FR}` is the Fisher-Rao distance and the penalty term is
    given by

    .. math::
        \mathcal{R}(\gamma) = \|\sqrt{\dot{\gamma}}- 1 \|_{\mathbb{L}^2}^2

    See the :footcite:`srivastava+klassen_2016_analysis_amplitude` for a
    detailed explanation.

    If the observations are defined in a :term:`domain` different than (0,1)
    their domains are normalized to this interval with an affine
    transformation.

    Args:
        fdata1: First FData object.
        fdata2: Second FData object.
        lam: Penalty term to restric the elasticity.
        eval_points: Array with points of evaluation.
        kwargs: Name arguments to be passed to
            :func:`elastic_registration_warping`.

    Returns:
        Elastic distance.

    Raises:
        ValueError: If the objects are not unidimensional.

    References:
        .. footbibliography::

    """
    from ...preprocessing.registration import FisherRaoElasticRegistration

    fdata1, fdata2 = _cast_to_grid(
        fdata1,
        fdata2,
        eval_points=eval_points,
        _check=_check,
    )

    # Both should have the same grid points
    eval_points_normalized = normalize_scale(fdata1.grid_points[0])

    # Calculate the corresponding srsf and normalize to (0,1)
    fdata1 = fdata1.copy(
        grid_points=eval_points_normalized,
        domain_range=(0, 1),
    )
    fdata2 = fdata2.copy(
        grid_points=eval_points_normalized,
        domain_range=(0, 1),
    )

    elastic_registration = FisherRaoElasticRegistration(
        template=fdata2,
        penalty=lam,
        output_points=eval_points_normalized,
        **kwargs,
    )

    fdata1_reg = elastic_registration.fit_transform(fdata1)

    srsf = SRSF(initial_value=0)
    fdata1_reg_srsf = srsf.fit_transform(fdata1_reg)
    fdata2_srsf = srsf.transform(fdata2)
    distance = l2_distance(fdata1_reg_srsf, fdata2_srsf)

    if lam != 0:
        # L2 norm || sqrt(Dh) - 1 ||^2
        warping_deriv = elastic_registration.warping_.derivative()
        penalty = warping_deriv(eval_points_normalized)[0, ..., 0]
        penalty = np.sqrt(penalty, out=penalty)
        penalty -= 1
        penalty = np.square(penalty, out=penalty)
        penalty = scipy.integrate.simps(penalty, x=eval_points_normalized)

        distance = np.sqrt(distance**2 + lam * penalty)

    return distance


def fisher_rao_phase_distance(
    fdata1: T,
    fdata2: T,
    *,
    lam: float = 0,
    eval_points: Optional[NDArrayFloat] = None,
    _check: bool = True,
) -> NDArrayFloat:
    r"""
    Compute the Fisher-Rao phase distance between two functional objects.

    Let :math:`f_i` and :math:`f_j` be two functional observations, and let
    :math:`\gamma_{ij}` the corresponding warping used in the elastic
    registration to align :math:`f_i` to :math:`f_j` (see
    :func:`elastic_registration`). The Fisher-Rao  phase distance between
    :math:`f_i` and :math:`f_j` is defined as

    .. math::
        d_{P}(f_i, f_j) = d_{FR}(\gamma_{ij}, \gamma_{id}) =
        arcos \left ( \int_0^1 \sqrt {\dot \gamma_{ij}(t)} dt \right )

    where :math:`\gamma_{id}` is the identity warping.

    See :footcite:`srivastava+klassen_2016_analysis_phase` for a detailed
    explanation.

    If the observations are defined in a :term:`domain` different than (0,1)
    their domains are normalized to this interval with an affine
    transformation.

    Args:
        fdata1: First FData object.
        fdata2: Second FData object.
        lam: Penalty term to restric the elasticity.
        eval_points (array_like, optional): Array with points of evaluation.

    Returns:
        Phase distance between the objects.

    Raises:
        ValueError: If the objects are not unidimensional.

    References:
        .. footbibliography::

    """
    from ...preprocessing.registration import FisherRaoElasticRegistration

    fdata1, fdata2 = _cast_to_grid(
        fdata1,
        fdata2,
        eval_points=eval_points,
        _check=_check,
    )

    # Rescale in the interval (0,1)
    eval_points_normalized = normalize_scale(fdata1.grid_points[0])

    # Calculate the corresponding srsf and normalize to (0,1)
    fdata1 = fdata1.copy(
        grid_points=eval_points_normalized,
        domain_range=(0, 1),
    )
    fdata2 = fdata2.copy(
        grid_points=eval_points_normalized,
        domain_range=(0, 1),
    )

    elastic_registration = FisherRaoElasticRegistration(
        penalty=lam,
        template=fdata2,
        output_points=eval_points_normalized,
    )

    elastic_registration.fit_transform(fdata1)

    warping_deriv = elastic_registration.warping_.derivative()
    derivative_warping = warping_deriv(eval_points_normalized)[0, ..., 0]

    derivative_warping = np.sqrt(derivative_warping, out=derivative_warping)

    d = scipy.integrate.simps(derivative_warping, x=eval_points_normalized)
    d = np.clip(d, -1, 1)

    return np.arccos(d)


def _fisher_rao_warping_distance(
    warping1: T,
    warping2: T,
    *,
    eval_points: Optional[NDArrayFloat] = None,
    _check: bool = True,
) -> NDArrayFloat:
    r"""
    Compute the Fisher-Rao distance between warpings functions.

    Let :math:`\gamma_i` and :math:`\gamma_j` be two warpings, defined in
    :math:`\gamma_i:[0,1] \rightarrow [0,1]`. The distance in the
    space of warping functions, :math:`\Gamma`, with the Riemannian metric
    given by the Fisher-Rao inner product can be computed using the structure
    of Hilbert sphere in their SRSF's.

    .. math::
        d_{\Gamma}(\gamma_i, \gamma_j) = cos^{-1} \left ( \int_0^1
        \sqrt{\dot \gamma_i(t)\dot \gamma_j(t)}dt \right )

    See :footcite:`srivastava+klassen_2016_analysis_probability` for a detailed
    explanation.

    If the warpings are not defined in [0,1], an affine transformation is made
    to change the :term:`domain`.

    Args:
        warping1: First warping.
        warping2: Second warping.
        eval_points: Array with points of evaluation.

    Returns:
        Distance between warpings:

    Raises:
        ValueError: If the objects are not unidimensional.

    References:
        .. footbibliography::

    """
    warping1, warping2 = _cast_to_grid(
        warping1,
        warping2,
        eval_points=eval_points,
        _check=_check,
    )

    # Normalization of warping to (0,1)x(0,1)
    warping1 = normalize_warping(warping1, (0, 1))
    warping2 = normalize_warping(warping2, (0, 1))

    warping1_data = warping1.derivative().data_matrix[0, ..., 0]
    warping2_data = warping2.derivative().data_matrix[0, ..., 0]

    # Derivative approximations can have negatives, specially in the
    # borders.
    warping1_data[warping1_data < 0] = 0
    warping2_data[warping2_data < 0] = 0

    # In this case the srsf is the sqrt(gamma')
    srsf_warping1 = np.sqrt(warping1_data, out=warping1_data)
    srsf_warping2 = np.sqrt(warping2_data, out=warping2_data)

    product = np.multiply(srsf_warping1, srsf_warping2, out=srsf_warping1)

    d = scipy.integrate.simps(product, x=warping1.grid_points[0])
    d = np.clip(d, -1, 1)

    return np.arccos(d)



from . import constants
from ._sklearn_adapter import (
    BaseEstimator,
    InductiveTransformerMixin,
    TransformerMixin,
)
from ._utils import (
    RandomStateLike,
    _cartesian_product,
    _check_array_key,
    _check_compatible_fdata,
    _check_compatible_fdatagrid,
    _check_estimator,
    _classifier_fit_depth_methods,
    _classifier_get_classes,
    _classifier_get_depth_methods,
    _compute_dependence,
    _DependenceMeasure,
    _evaluate_grid,
    _fit_feature_transformer,
    _int_to_real,
    _pairwise_symmetric,
    _reshape_eval_points,
    _same_domain,
    _to_array_maybe_ragged,
    _to_domain_range,
    _to_grid,
    _to_grid_points,
    check_is_univariate,
    nquad_vec,
)
from ._warping import invert_warping, normalize_scale, normalize_warping



from __future__ import annotations

from typing import Optional, Union, overload

import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from sklearn.utils.validation import check_is_fitted
from typing_extensions import Literal

from ._functional_data import FData
from ._typing import ArrayLike, GridPointsLike
from .extrapolation import ExtrapolationLike
from .grid import FDataGrid


class EvaluationTransformer(
    BaseEstimator,  # type:ignore
    TransformerMixin,  # type:ignore
):
    r"""
    Transformer returning the evaluations of FData objects as a matrix.

    Args:
        eval_points (array_like): List of points where the functions are
            evaluated. If `None`, the functions must be `FDatagrid` objects
            and all points will be returned.
        extrapolation (str or Extrapolation, optional): Controls the
            extrapolation mode for elements outside the :term:`domain` range.
            By default it is used the mode defined during the instance of the
            object.
        grid (bool, optional): Whether to evaluate the results on a grid
            spanned by the input arrays, or at points specified by the
            input arrays. If true the eval_points should be a list of size
            dim_domain with the corresponding times for each axis. The
            return matrix has shape n_samples x len(t1) x len(t2) x ... x
            len(t_dim_domain) x dim_codomain. If the domain dimension is 1
            the parameter has no efect. Defaults to False.

    Attributes:
        shape\_ (tuple): original shape of coefficients per sample.

    Examples:
        >>> from skfda.representation import (FDataGrid, FDataBasis,
        ...                                   EvaluationTransformer)
         >>> from skfda.representation.basis import Monomial

        Functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}\longmapsto\mathbb{R}`.

        >>> data_matrix = [[1, 2], [2, 3]]
        >>> grid_points = [2, 4]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>>
        >>> transformer = EvaluationTransformer()
        >>> transformer.fit_transform(fd)
        array([[ 1., 2.],
               [ 2., 3.]])

        Functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}\longmapsto\mathbb{R}^2`.

        >>> data_matrix = [[[1, 0.3], [2, 0.4]], [[2, 0.5], [3, 0.6]]]
        >>> grid_points = [2, 4]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>>
        >>> transformer = EvaluationTransformer()
        >>> transformer.fit_transform(fd)
        array([[ 1. ,  0.3,  2. ,  0.4],
               [ 2. ,  0.5,  3. ,  0.6]])

        Representation of a functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}^2\longmapsto\mathbb{R}`.

        >>> data_matrix = [[[1, 0.3], [2, 0.4]], [[2, 0.5], [3, 0.6]]]
        >>> grid_points = [[2, 4], [3, 6]]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>>
        >>> transformer = EvaluationTransformer()
        >>> transformer.fit_transform(fd)
        array([[ 1. ,  0.3,  2. ,  0.4],
               [ 2. ,  0.5,  3. ,  0.6]])

        Evaluation of a functional data object at several points.

        >>> basis = Monomial(n_basis=4)
        >>> coefficients = [[0.5, 1, 2, .5], [1.5, 1, 4, .5]]
        >>> fd = FDataBasis(basis, coefficients)
        >>>
        >>> transformer = EvaluationTransformer([0, 0.2, 0.5, 0.7, 1])
        >>> transformer.fit_transform(fd)
        array([[ 0.5   ,  0.784 ,  1.5625,  2.3515,  4.    ],
               [ 1.5   ,  1.864 ,  3.0625,  4.3315,  7.    ]])

    """

    @overload
    def __init__(
        self,
        eval_points: ArrayLike,
        *,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[False] = False,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        eval_points: GridPointsLike,
        *,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[True],
    ) -> None:
        pass

    def __init__(
        self,
        eval_points: Union[ArrayLike, GridPointsLike, None] = None,
        *,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: bool = False,
    ):
        self.eval_points = eval_points
        self.extrapolation = extrapolation
        self.grid = grid

    def fit(  # noqa: D102
        self,
        X: FData,
        y: None = None,
    ) -> EvaluationTransformer:

        if self.eval_points is None and not isinstance(X, FDataGrid):
            raise ValueError(
                "If no eval_points are passed, the functions "
                "should be FDataGrid objects.",
            )

        self._is_fitted = True

        return self

    def transform(  # noqa: D102
        self,
        X: FData,
        y: None = None,
    ) -> np.ndarray:

        check_is_fitted(self, '_is_fitted')

        if self.eval_points is None:
            evaluation = X.data_matrix.copy()
        else:
            evaluation = X(  # type: ignore
                self.eval_points,
                extrapolation=self.extrapolation,
                grid=self.grid,
            )

        return evaluation.reshape((X.n_samples, -1))



"""Depth and outlyingness ABCs and implementations for multivariate data."""

from __future__ import annotations

import abc
import math
from typing import Generic, Optional, TypeVar

import numpy as np
import scipy.stats
import sklearn
from scipy.special import comb
from typing_extensions import Literal

T = TypeVar("T", contravariant=True)
SelfType = TypeVar("SelfType")
_Side = Literal["left", "right"]


class _DepthOrOutlyingness(
    abc.ABC,
    sklearn.base.BaseEstimator,  # type: ignore
    Generic[T],
):
    """Abstract class representing a depth or outlyingness function."""

    def fit(self: SelfType, X: T, y: None = None) -> SelfType:
        """
        Learn the distribution from the observations.

        Args:
            X: Functional dataset from which the distribution of the data is
                inferred.
            y: Unused. Kept only for convention.

        Returns:
            Fitted estimator.

        """
        return self

    @abc.abstractmethod
    def transform(self, X: T) -> np.ndarray:
        """
        Compute the depth or outlyingness inside the learned distribution.

        Args:
            X: Points whose depth is going to be evaluated.

        Returns:
            Depth of each observation.

        """
        pass

    def fit_transform(self, X: T, y: None = None) -> np.ndarray:
        """
        Compute the depth or outlyingness of each observation.

        This computation is done with respect to the whole dataset.

        Args:
            X: Dataset.
            y: Unused. Kept only for convention.

        Returns:
            Depth of each observation.

        """
        return self.fit(X).transform(X)

    def __call__(
        self,
        X: T,
        *,
        distribution: Optional[T] = None,
    ) -> np.ndarray:
        """
        Allow the depth or outlyingness to be used as a function.

        Args:
            X: Points whose depth is going to be evaluated.
            distribution: Functional dataset from which the distribution of
                the data is inferred. If ``None`` it is the same as ``X``.

        Returns:
            Depth of each observation.

        """
        copy = sklearn.base.clone(self)

        if distribution is None:
            return copy.fit_transform(X)

        return copy.fit(distribution).transform(X)

    @property  # noqa: WPS125
    def max(self) -> float:  # noqa: WPS125
        """
        Maximum (or supremum if there is no maximum) of the possibly predicted
        values.

        """
        return 1

    @property  # noqa: WPS125
    def min(self) -> float:  # noqa: WPS125
        """
        Minimum (or infimum if there is no maximum) of the possibly predicted
        values.

        """
        return 0


class Depth(_DepthOrOutlyingness[T]):
    """Abstract class representing a depth function."""


class Outlyingness(_DepthOrOutlyingness[T]):
    """Abstract class representing an outlyingness function."""


def _searchsorted_one_dim(
    array: np.ndarray,
    values: np.ndarray,
    *,
    side: _Side = 'left',
) -> np.ndarray:
    return np.searchsorted(array, values, side=side)


_searchsorted_vectorized = np.vectorize(
    _searchsorted_one_dim,
    signature='(n),(m),()->(m)',
    excluded='side',
)


def _searchsorted_ordered(
    array: np.ndarray,
    values: np.ndarray,
    *,
    side: _Side = 'left',
) -> np.ndarray:
    return _searchsorted_vectorized(array, values, side=side)


def _cumulative_distribution(column: np.ndarray) -> np.ndarray:
    """Calculate the cumulative distribution function at each point.

    Args:
        column (numpy.darray): Array containing the values over which the
            distribution function is calculated.

    Returns:
        numpy.darray: Array containing the evaluation at each point of the
            distribution function.

    Examples:
        >>> _cumulative_distribution(np.array([1, 4, 5, 1, 2, 2, 4, 1, 1, 3]))
        array([ 0.4,  0.9,  1. ,  0.4,  0.6,  0.6,  0.9,  0.4,  0.4,  0.7])

    """
    return _searchsorted_ordered(
        np.sort(column),
        column,
        side='right',
    ) / len(column)


class _UnivariateFraimanMuniz(Depth[np.ndarray]):
    r"""
    Univariate depth used to compute the Fraiman an Muniz depth.

    Each column is considered as the samples of an aleatory variable.
    The univariate depth of each of the samples of each column is calculated
    as follows:

    .. math::
        D(x) = 1 - \left\lvert \frac{1}{2}- F(x)\right\rvert

    Where :math:`F` stands for the marginal univariate distribution function of
    each column.

    """

    def fit(self: SelfType, X: np.ndarray, y: None = None) -> SelfType:
        self._sorted_values = np.sort(X, axis=0)
        return self

    def transform(self, X: np.ndarray) -> np.ndarray:
        cum_dist = _searchsorted_ordered(
            np.moveaxis(self._sorted_values, 0, -1),
            np.moveaxis(X, 0, -1),
            side='right',
        ) / len(self._sorted_values)

        assert cum_dist.shape[-2] == 1
        return 1 - np.abs(0.5 - np.moveaxis(cum_dist, -1, 0)[..., 0])

    @property  # noqa: WPS125
    def min(self) -> float:  # noqa: WPS125
        return 1 / 2


class SimplicialDepth(Depth[np.ndarray]):
    r"""
    Simplicial depth.

    The simplicial depth of a point :math:`x` in :math:`\mathbb{R}^p` given a
    distribution :math:`F` is the probability that a random simplex with its
    :math:`p + 1` points sampled from :math:`F` contains :math:`x`.

    References:
        Liu, R. Y. (1990). On a Notion of Data Depth Based on Random
        Simplices. The Annals of Statistics, 18(1), 405–414.


    """

    def fit(  # noqa: D102
        self,
        X: np.ndarray,
        y: None = None,
    ) -> SimplicialDepth:
        self._dim = X.shape[-1]

        if self._dim == 1:
            self.sorted_values = np.sort(X, axis=0)
        else:
            raise NotImplementedError(
                "SimplicialDepth is currently only "
                "implemented for one-dimensional data.",
            )

        return self

    def transform(self, X: np.ndarray) -> np.ndarray:  # noqa: D102

        assert self._dim == X.shape[-1]

        if self._dim == 1:
            positions_left = _searchsorted_ordered(
                np.moveaxis(self.sorted_values, 0, -1),
                np.moveaxis(X, 0, -1),
            )

            positions_left = np.moveaxis(positions_left, -1, 0)[..., 0]

            positions_right = _searchsorted_ordered(
                np.moveaxis(self.sorted_values, 0, -1),
                np.moveaxis(X, 0, -1),
                side='right',
            )

            positions_right = np.moveaxis(positions_right, -1, 0)[..., 0]

            num_strictly_below = positions_left
            num_strictly_above = len(self.sorted_values) - positions_right

            total_pairs = comb(len(self.sorted_values), 2)

        return (
            total_pairs - comb(num_strictly_below, 2)
            - comb(num_strictly_above, 2)
        ) / total_pairs


class OutlyingnessBasedDepth(Depth[T]):
    r"""
    Computes depth based on an outlyingness measure.

    An outlyingness function :math:`O(x)` can be converted to a depth
    function as

    .. math::
        D(x) = \frac{1}{1 + O(x)}

    if :math:`O(x)` is unbounded or as

    .. math::
        D(x) = 1 - \frac{O(x)}{\sup O(x)}

    if :math:`O(x)` is bounded. If the infimum value of the
    outlyiness function is not zero, it is subtracted beforehand.

    Args:
        outlyingness (Outlyingness): Outlyingness object.

    References:
        Serfling, R. (2006). Depth functions in nonparametric
        multivariate inference. DIMACS Series in Discrete Mathematics and
        Theoretical Computer Science, 72, 1.

    """

    def __init__(self, outlyingness: Outlyingness[T]):
        self.outlyingness = outlyingness

    def fit(  # noqa: D102
        self,
        X: T,
        y: None = None,
    ) -> OutlyingnessBasedDepth[T]:
        self.outlyingness.fit(X)

        return self

    def transform(self, X: np.ndarray) -> np.ndarray:  # noqa: D102
        outlyingness_values = self.outlyingness.transform(X)

        min_val = self.outlyingness.min
        max_val = self.outlyingness.max

        if math.isinf(max_val):
            return 1 / (1 + outlyingness_values - min_val)

        return 1 - (outlyingness_values - min_val) / (max_val - min_val)


class StahelDonohoOutlyingness(Outlyingness[np.ndarray]):
    r"""
    Computes Stahel-Donoho outlyingness.

    Stahel-Donoho outlyingness is defined as

    .. math::
        \sup_{\|u\|=1} \frac{|u^T x - \text{Med}(u^T X))|}{\text{MAD}(u^TX)}

    where :math:`\text{X}` is a sample with distribution :math:`F`,
    :math:`\text{Med}` is the median and :math:`\text{MAD}` is the
    median absolute deviation.

    References:
        Zuo, Y., Cui, H., & He, X. (2004). On the Stahel-Donoho
        estimator and depth-weighted means of multivariate data. Annals of
        Statistics, 32(1), 167–188. https://doi.org/10.1214/aos/1079120132

    """

    def fit(  # noqa: D102
        self,
        X: np.ndarray,
        y: None = None,
    ) -> StahelDonohoOutlyingness:

        dim = X.shape[-1]

        if dim == 1:
            self._location = np.median(X, axis=0)
            self._scale = scipy.stats.median_abs_deviation(X, axis=0)
        else:
            raise NotImplementedError("Only implemented for one dimension")

        return self

    def transform(self, X: np.ndarray) -> np.ndarray:  # noqa: D102

        dim = X.shape[-1]

        if dim == 1:
            # Special case, can be computed exactly
            return (
                np.abs(X - self._location)
                / self._scale
            )[..., 0]

        raise NotImplementedError("Only implemented for one dimension")

    @property  # noqa: WPS125
    def max(self) -> float:  # noqa: WPS125
        return math.inf


class ProjectionDepth(OutlyingnessBasedDepth[np.ndarray]):
    r"""
    Computes Projection depth.

    It is defined as the depth induced by the
    :class:`Stahel-Donoho outlyingness <StahelDonohoOutlyingness>`.

    See also:
        :class:`StahelDonohoOutlyingness`: Stahel-Donoho outlyingness.

    References:
        Zuo, Y., Cui, H., & He, X. (2004). On the Stahel-Donoho
        estimator and depth-weighted means of multivariate data. Annals of
        Statistics, 32(1), 167–188. https://doi.org/10.1214/aos/1079120132

    """

    def __init__(self) -> None:
        super().__init__(outlyingness=StahelDonohoOutlyingness())



import io
import math
import re
from itertools import repeat
from typing import Optional, Sequence, Tuple, TypeVar, Union

import matplotlib.backends.backend_svg
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from typing_extensions import Protocol

from ...representation._functional_data import FData

non_close_text = '[^>]*?'
svg_width_regex = re.compile(
    f'(<svg {non_close_text}width="){non_close_text}("{non_close_text}>)',
)
svg_width_replacement = r'\g<1>100%\g<2>'
svg_height_regex = re.compile(
    f'(<svg {non_close_text})height="{non_close_text}"({non_close_text}>)',
)
svg_height_replacement = r'\g<1>\g<2>'

ColorLike = Union[
    Tuple[float, float, float],
    Tuple[float, float, float, float],
    str,
    Sequence[float],
]

K = TypeVar('K', contravariant=True)
V = TypeVar('V', covariant=True)


class Indexable(Protocol[K, V]):
    """Class Indexable used to type _get_color_info."""

    def __getitem__(self, __key: K) -> V:
        pass

    def __len__(self) -> int:
        pass


def _create_figure() -> Figure:
    """Create figure using the default backend."""
    return plt.figure()


def _figure_to_svg(figure: Figure) -> str:
    """Return the SVG representation of a figure."""
    old_canvas = figure.canvas
    matplotlib.backends.backend_svg.FigureCanvas(figure)
    output = io.BytesIO()
    figure.savefig(output, format='svg')
    figure.set_canvas(old_canvas)
    data = output.getvalue()
    decoded_data = data.decode('utf-8')

    new_data = svg_width_regex.sub(
        svg_width_replacement,
        decoded_data,
        count=1,
    )

    return svg_height_regex.sub(
        svg_height_replacement,
        new_data,
        count=1,
    )


def _get_figure_and_axes(
    chart: Union[Figure, Axes, Sequence[Axes], None] = None,
    fig: Optional[Figure] = None,
    axes: Union[Axes, Sequence[Axes], None] = None,
) -> Tuple[Figure, Sequence[Axes]]:
    """Obtain the figure and axes from the arguments."""
    num_defined = sum(e is not None for e in (chart, fig, axes))
    if num_defined > 1:
        raise ValueError(
            "Only one of chart, fig and axes parameters"
            "can be passed as an argument.",
        )

    # Parse chart argument
    if chart is not None:
        if isinstance(chart, matplotlib.figure.Figure):
            fig = chart
        else:
            axes = chart

    if fig is None and axes is None:
        new_fig = _create_figure()
        new_axes = []

    elif fig is not None:
        new_fig = fig
        new_axes = fig.axes

    else:
        assert axes is not None
        if isinstance(axes, Axes):
            axes = [axes]

        new_fig = axes[0].figure
        new_axes = axes

    return new_fig, new_axes


def _get_axes_shape(
    n_axes: int,
    n_rows: Optional[int] = None,
    n_cols: Optional[int] = None,
) -> Tuple[int, int]:
    """Get the number of rows and columns of the subplots."""
    if (
        (n_rows is not None and n_cols is not None)
        and ((n_rows * n_cols) < n_axes)
    ):
        raise ValueError(
            f"The number of rows ({n_rows}) multiplied by "
            f"the number of columns ({n_cols}) "
            f"is less than the number of required "
            f"axes ({n_axes})",
        )

    if n_rows is None and n_cols is None:
        new_n_cols = int(math.ceil(math.sqrt(n_axes)))
        new_n_rows = int(math.ceil(n_axes / new_n_cols))
    elif n_rows is None and n_cols is not None:
        new_n_cols = n_cols
        new_n_rows = int(math.ceil(n_axes / n_cols))
    elif n_cols is None and n_rows is not None:
        new_n_cols = int(math.ceil(n_axes / n_rows))
        new_n_rows = n_rows

    return new_n_rows, new_n_cols


def _projection_from_dim(dim: int) -> str:

    if dim == 2:
        return 'rectilinear'
    elif dim == 3:
        return '3d'

    raise NotImplementedError(
        "Only bidimensional or tridimensional plots are supported.",
    )


def _set_figure_layout(
    fig: Figure,
    axes: Sequence[Axes],
    dim: Union[int, Sequence[int]] = 2,
    n_axes: int = 1,
    n_rows: Optional[int] = None,
    n_cols: Optional[int] = None,
) -> Tuple[Figure, Sequence[Axes]]:
    """
    Set the figure axes for plotting.

    Args:
        fig: Figure over with the graphs are plotted in case ax is not
            specified.
        axes: Axis over where the graphs are plotted.
        dim: Dimension of the plot. Either 2 for a 2D plot or 3 for a 3D plot.
        n_axes: Number of subplots.
        n_rows: Designates the number of rows of the figure to plot the
            different dimensions of the image. Can only be passed if no axes
            are specified.
        n_cols: Designates the number of columns of the figure to plot the
            different dimensions of the image. Can only be passed if no axes
            are specified.

    Returns:
        (tuple): tuple containing:

            * fig (figure): figure object in which the graphs are plotted.
            * axes (list): axes in which the graphs are plotted.

    """
    if len(axes) not in {0, n_axes}:
        raise ValueError(
            f"The number of axes ({len(axes)}) must be 0 (to create them)"
            f" or equal to the number of axes needed "
            f"({n_axes} in this case).",
        )

    if len(axes) != 0 and (n_rows is not None or n_cols is not None):
        raise ValueError(
            "The number of columns and/or number of rows of "
            "the figure, in which each dimension of the "
            "image is plotted, can only be customized in case "
            "that no axes are provided.",
        )

    if len(axes) == 0:
        # Create the axes

        n_rows, n_cols = _get_axes_shape(n_axes, n_rows, n_cols)

        for i in range(n_rows):
            for j in range(n_cols):
                subplot_index = i * n_cols + j
                if subplot_index < n_axes:
                    plot_dim = (
                        dim if isinstance(dim, int) else dim[subplot_index]
                    )

                fig.add_subplot(
                    n_rows,
                    n_cols,
                    subplot_index + 1,
                    projection=_projection_from_dim(plot_dim),
                )

        axes = fig.axes

    else:
        # Check that the projections are right
        projections = (
            repeat(_projection_from_dim(dim))
            if isinstance(dim, int)
            else (_projection_from_dim(d) for d in dim)
        )

        for a, proj in zip(axes, projections):
            if a.name != proj:
                raise ValueError(
                    f"The projection of the axes is {a.name} "
                    f"but should be {proj}",
                )

    return fig, axes


def _set_labels(
    fdata: FData,
    fig: Figure,
    axes: Sequence[Axes],
    patches: Optional[Sequence[matplotlib.patches.Patch]] = None,
) -> None:
    """Set labels if any.

    Args:
        fdata: functional data object.
        fig: figure object containing the axes that implement
            set_xlabel and set_ylabel, and set_zlabel in case
            of a 3d projection.
        axes: axes objects that implement set_xlabel and set_ylabel,
            and set_zlabel in case of a 3d projection; used if
            fig is None.
        patches: objects used to generate each entry in the legend.

    """
    # Dataset name
    if fdata.dataset_name is not None:
        fig.suptitle(fdata.dataset_name)

    # Legend
    if patches is not None:
        fig.legend(handles=patches)
    elif patches is not None:
        axes[0].legend(handles=patches)

    assert len(axes) == fdata.dim_codomain

    # Axis labels
    if axes[0].name == '3d':
        for i, a in enumerate(axes):
            if fdata.argument_names[0] is not None:
                a.set_xlabel(fdata.argument_names[0])
            if fdata.argument_names[1] is not None:
                a.set_ylabel(fdata.argument_names[1])
            if fdata.coordinate_names[i] is not None:
                a.set_zlabel(fdata.coordinate_names[i])
    else:
        for i in range(fdata.dim_codomain):
            if fdata.argument_names[0] is not None:
                axes[i].set_xlabel(fdata.argument_names[0])
            if fdata.coordinate_names[i] is not None:
                axes[i].set_ylabel(fdata.coordinate_names[i])


def _change_luminosity(color: ColorLike, amount: float = 0.5) -> ColorLike:
    """
    Change the given color luminosity by the given amount.

    Input can be matplotlib color string, hex string, or RGB tuple.

    Note:
        Based on https://stackoverflow.com/a/49601444/2455333
    """
    import colorsys

    import matplotlib.colors as mc
    try:
        c = mc.cnames[color]
    except TypeError:
        c = color
    c = colorsys.rgb_to_hls(*mc.to_rgb(c))

    intensity = (amount - 0.5) * 2
    up = intensity > 0
    intensity = abs(intensity)

    lightness = c[1]
    if up:
        new_lightness = lightness + intensity * (1 - lightness)
    else:
        new_lightness = lightness - intensity * lightness

    return colorsys.hls_to_rgb(c[0], new_lightness, c[2])


def _darken(color: ColorLike, amount: float = 0) -> ColorLike:
    return _change_luminosity(color, 0.5 - amount / 2)


def _lighten(color: ColorLike, amount: float = 0) -> ColorLike:
    return _change_luminosity(color, 0.5 + amount / 2)



"""Unsupervised learner for implementing neighbor searches."""

from .._neighbors_base import (
    KNeighborsMixin,
    NeighborsBase,
    NeighborsMixin,
    RadiusNeighborsMixin,
)


class NearestNeighbors(NeighborsBase, NeighborsMixin, KNeighborsMixin,
                       RadiusNeighborsMixin):
    """Unsupervised learner for implementing neighbor searches.

    Parameters
    ----------
    n_neighbors : int, optional (default = 5)
        Number of neighbors to use by default for :meth:`kneighbors` queries.
    radius : float, optional (default = 1.0)
        Range of parameter space to use by default for :meth:`radius_neighbors`
        queries.
    algorithm : {'auto', 'ball_tree', 'brute'}, optional
        Algorithm used to compute the nearest neighbors:

        - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`.
        - 'brute' will use a brute-force search.
        - 'auto' will attempt to decide the most appropriate algorithm based on
          the values passed to :meth:`fit` method.

    leaf_size : int, optional (default = 30)
        Leaf size passed to BallTree or KDTree.  This can affect the
        speed of the construction and query, as well as the memory
        required to store the tree.  The optimal value depends on the
        nature of the problem.
    metric : string or callable, (default
        :func:`l2_distance <skfda.misc.metrics.l2_distance>`)
        the distance metric to use for the tree.  The default metric is
        the L2 distance. See the documentation of the metrics module
        for a list of available metrics.
    metric_params : dict, optional (default = None)
        Additional keyword arguments for the metric function.
    n_jobs : int or None, optional (default=None)
        The number of parallel jobs to run for neighbors search.
        ``None`` means 1 unless in a :obj:`joblib.parallel_backend` context.
        ``-1`` means using all processors.
        Doesn't affect :meth:`fit` method.
    multivariate_metric : boolean, optional (default = False)
        Indicates if the metric used is a sklearn distance between vectors (see
        :class:`sklearn.neighbors.DistanceMetric`) or a functional metric of
        the module :mod:`skfda.misc.metrics`.
    Examples
    --------
    Firstly, we will create a toy dataset with 2 classes

    >>> from skfda.datasets import make_sinusoidal_process
    >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
    >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
    ...                               phase_std=.25, random_state=0)
    >>> fd = fd1.concatenate(fd2)

    We will fit a Nearest Neighbors estimator

    >>> from skfda.ml.clustering import NearestNeighbors
    >>> neigh = NearestNeighbors(radius=.3)
    >>> neigh.fit(fd)
    NearestNeighbors(...radius=0.3...)

    Now we can query the k-nearest neighbors.

    >>> distances, index = neigh.kneighbors(fd[:2])
    >>> index # Index of k-neighbors of samples 0 and 1
    array([[ 0,  7,  6, 11,  2],...)

    >>> distances.round(2) # Distances to k-neighbors
    array([[ 0.  ,  0.28,  0.29,  0.29,  0.3 ],
           [ 0.  ,  0.27,  0.28,  0.29,  0.3 ]])

    We can query the neighbors in a given radius too.

    >>> distances, index = neigh.radius_neighbors(fd[:2])
    >>> index[0]
    array([ 0,  2,  6,  7, 11]...)

    >>> distances[0].round(2) # Distances to neighbors of the sample 0
    array([ 0.  ,  0.3 ,  0.29,  0.28,  0.29])

    See also
    --------
    :class:`~skfda.ml.classification.KNeighborsClassifier`
    :class:`~skfda.ml.classification.RadiusNeighborsClassifier`
    :class:`~skfda.ml.classification.NearestCentroids`
    :class:`~skfda.ml.regression.KNeighborsRegressor`
    :class:`~skfda.ml.regression.RadiusNeighborsRegressor`


    Notes
    -----
    See Nearest Neighbors in the sklearn online documentation for a discussion
    of the choice of ``algorithm`` and ``leaf_size``.

    This class wraps the sklearn classifier
    `sklearn.neighbors.KNeighborsClassifier`.

    https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm

    """

    def __init__(self, n_neighbors=5, radius=1.0, algorithm='auto',
                 leaf_size=30, metric='l2', metric_params=None,
                 n_jobs=1, multivariate_metric=False):
        """Initialize the nearest neighbors searcher."""

        super().__init__(n_neighbors=n_neighbors, radius=radius,
                         algorithm=algorithm, leaf_size=leaf_size,
                         metric=metric, metric_params=metric_params,
                         n_jobs=n_jobs,
                         multivariate_metric=multivariate_metric)

    def _init_estimator(self, sklearn_metric):
        """Initialize the sklearn nearest neighbors estimator.

        Args:
            sklearn_metric: (pyfunc or 'precomputed'): Metric compatible with
                sklearn API or matrix (n_samples, n_samples) with precomputed
                distances.

        Returns:
            Sklearn K Neighbors estimator initialized.

        """
        from sklearn.neighbors import NearestNeighbors as _NearestNeighbors

        return _NearestNeighbors(
            n_neighbors=self.n_neighbors, radius=self.radius,
            algorithm=self.algorithm, leaf_size=self.leaf_size,
            metric=sklearn_metric, metric_params=self.metric_params,
            n_jobs=self.n_jobs)
